"""File: choose.py
:note: called from distribution.py by Producer.calc_info
Provides functions for choice of production activity
(determining pclass). Many functions return a dict,
with the following keys:

    pclass: int 
        production class (0, 1, 2, 3, 4)
    tr_d : float 
        quantity of leisure
    tw_d : float
        time working off-farm
    th_d : float 
        quantity of time working on-farm
    ts_d : float 
        quantity of time working on-farm
    L_d : float 
        quantity of hired labor
    h_d : float
        quantity of land operated
    q_d : float
        output planned.
    Y_d : float
        Income
    U_d : float 
        Utility at the maximum
"""
import math, logging
import numpy as np
from scipy.optimize import fminbound

### global constant ###
constraints = dict(
    B=(1<<0),
    R1=(1<<1),
    rA=(1<<2),
    ell0=(1<<3),
    )



def cultivate(agent, K, v, w, rho, params):
    """Return dict of agent attributes,
    including pclass, factor demands, and expected welfare.
    
    Parameters
    ----------
    agent : Producer
        (Need its hbar and bbar.)
    K : float
        sector specific fixed cost (KT or KM)
    v : float 
        Price of land
    w : float
        Price of labor
    rho : float 
        Premium associated with participation in modern value chain
    params : dict
        Model parameters, especially s1, s2, D, KM, KT.
        (See distribution.py for the definitions)

    Returns
    -------
    agentinfo : dict
        keys are pclass,tr_d,tw_d,th_d,L_d,h_d,q_d,Y_d,lmda,U_d
        (See the module documentation for definitions.)
    """
    assert K in (params['KT'],params['KM']), 'unrecognized fixed cost type'
    assert rho >= 1, "rho={}; caller shd ensure rho >= 1".format(rho)
    B = agent.bbar + v*agent.hbar - K  #working capital depends on hbar and K
    if math.isnan(B):
        msg = "B={}, bbar={}, hbar={}, K={}, v={}, w={}, rho={}"
        msg = msg.format(B,agent.bbar,agent.hbar,K,v,w,rho)
        raise ValueError(msg)
    B1, B2, B3 = b123(agent, K, v, w, rho, params)
    if B < B1:
        # Producer is LC (laborer_cultivator )
        agentinfo = laborer_cultivator(agent, B, K, v, w, rho, params)
    else:
        # Producer is SC, SM, or LG
        agentinfo = cultivator(agent, B, K, v, w, rho, params)
    agentinfo['K'] = K
    return agentinfo


def b123(agent, K, v, w, rho, params):
    """Return 3-tuple of float,
    the values of B1, B2, and B3.
    """
    assert (K==params['KT'] or K==params['KM'])
    A = agent.A       #Productivity parameter
    s1 = params['s1'] #first-order aspect of supervision function
    s2 = params['s2'] #second-order aspect of supervision function
    D = params['D']   #scale factor: sub-utility of leisure
    dar = D / (rho * A)
    dar2 = dar * dar
    dar4 = dar2 * dar2
    B1 = w - v*dar*dar
    B2 = B1 + w* s1/(1 - s1)
    a2r2 = A*A*rho*rho
    assert np.allclose(B2, w/(1 - s1) - v*dar*dar)
    old_disc = (4*s2**2*D**4*v**2 - 4*a2r2*s2*D**2*v*w + a2r2*a2r2*(s1**2 + 4*s2)*w**2)
    oldB3 = -(2*s1*s2*D**2*v - 2*s1**2*s2*D**2*v - 8*s2**2*D**2*v \
        + a2r2*(-2*s1*w + s1**2*w + s1**3*w + 4*s2*w + 4*s1*s2*w) \
        - (-2 + s1 + s1**2 + 4*s2) \
        * np.sqrt(4*s2**2*D**4*v**2 - 4*a2r2*s2*D**2*v*w + a2r2*a2r2*(s1**2 + 4*s2)*w**2))\
        /(2*a2r2*s2*(-1 + s1**2 + 4*s2))
    disc = (s2*s2*dar4*v*v - s2*dar2*v*w + (s1**2/4 + s2)*w*w)
    assert np.allclose(old_disc, 4*a2r2*a2r2 * disc)
    if disc < 0: #shd only happen due to rounding error
        assert np.allclose(0, disc), "{}".format(disc)
        disc = 0
    B3 = ((s1 - s1*s1*(1 + s1)/2 - 2*s2 - 2*s1*s2)*w \
        + (s1 + s1**2 + 4*s2 - 2) * np.sqrt(disc) \
        - dar2*v*s2*(s1 - s1**2 - 4*s2)) \
       /(s2*(-1 + s1**2 + 4*s2))
    if disc > 0:
        assert np.allclose(B3,oldB3)
    assert B1 <= B2 <= B3, 'B1:{B1}, B2:{B2}, B3:{B3}'.format(B1=B1,B2=B2,B3=B3)
    return B1, B2, B3


def cultivator(agent, B, K, v, w, rho, params):
    """Return dict of producer attributes,
    including pclass, factor demands, and welfare.
    1. Try SM. If th<0, switch to LG. Elif L<0 switch to SC.
    2. If capital constraint binds, delegate to constrained_cultivator.

    Parameters
    ----------
    agent : Producer
        the agent (we need its hbar and A attributes)
    K : float 
        Fixed set-up costs 
    rho : float 
        Premium associated with participation in modern value chain
    v : float 
        Price of land
    w : float
        Price of labor
    params : dict
        Model parameters, especially s1, s2, D, KM, KT.
        (see distribution.py for the definitions)

    Returns
    -------
    result : dict
        keys are pclass,tr_d,tw_d,th_d,L_d,h_d,q_d,Y_d,lmda,U_d
        (See the module documentation for definitions.)

    Warning:
    --------
    Currently has multiple (3) return points.
    """
    assert (K==params['KT'] or K==params['KM'])
    hbar, A = agent.hbar, agent.A
    s1, s2, D = params['s1'], params['s2'], params['D']
    result = dict(tw_d=0, constraints=0)
    dar = D / (A * rho)
    dd2  = D*D
    aa2 = A*A
    rr2 = rho*rho
    vv2 = v*v
    a2r2 = aa2*rr2
    ar4vw01 = a2r2 - 4*v*w
    ar4vw02 = a2r2 / (4 * v * w)
    #compute SMU01
    pclass = 3   #SM
    L_u = (1-s1-1/ar4vw02)/(2*s2)
    #
    assert np.allclose(L_u, (1-s1)/(2*s2) - (2 * v * w)/(s2*a2r2) )
    #
    tr_u = (4*dar*dar*vv2)/a2r2
    ts_u = s1*L_u + s2*L_u*L_u
    th_u = 1 - tr_u - ts_u
    h_u = (w/v)*ar4vw02*(1+L_u-ts_u-tr_u)
    if (L_u < 0): #try SCU instead (i.e., impose L=0)
        assert (ar4vw02 < 1/(1-s1))
        if (a2r2 <= 2 * D * v): #production does not pay
            result = pure_laborer(agent, v, w, params)
            result['Y_d'] -= K
            result['U_d'] -= K
            return result
            #^^^^^ #:todo: ugly return point
        else: #SCU
            pclass = 2
            L_u = 0
            ts_u = 0
            h_u = a2r2/(4*vv2) - dar*dar
            tr_u = 4*v*v*dar*dar/a2r2
            th_u = 1 - tr_u
            assert np.allclose(th_u, h_u*4*vv2/a2r2)
    elif (th_u < 0): #try LGU instead
        assert (ar4vw01 >= 2*D*v*s1)  #weighted average of above conditions -> L>=0 for LGU
        pclass = 4
        th_u = 0
        s12 = s1/(2*s2)
        L_u = -s12 + math.sqrt(s12*s12 + 1/s2) * ar4vw01 / math.sqrt(16*dd2*vv2*s2+ar4vw01*ar4vw01)
        ts_u = s1*L_u + s2*L_u*L_u
        assert L_u>=0, 'L_u={} (should have L_u>0 if it pays to produce)'.format(L_u)
        h_u = (a2r2/vv2) * (L_u / 4)
        tr_u = 1 - ts_u
        assert tr_u >= 0, 'L_u={0} should be constrained to {1}'.format(L_u, maxL)
    else: #SMU applies
        assert np.allclose(h_u, (w/v)*(1+(1-s1)*L_u-s2*L_u*L_u)/(1-s1-2*s2*L_u)-dar*dar),\
        "{} vs. {}".format(h_u,  (w/v)*(1+(1-s1)*L_u-s2*L_u*L_u)/(1-s1-2*s2*L_u)-dar*dar)
    #test for capital constraint
    if (v*h_u + w*L_u <= B): #does not bind! we're done!
        assert (0 <= tr_u <= 1)
        msg = "cultivator: for pclass {}, capital constraint does not bind".format(pclass)
        logging.debug(msg)
        q_u = A * math.sqrt(h_u*L_u)
        Y_u = rho*q_u - v*(h_u - hbar) - w*L_u - K
        U_u = Y_u + D*math.sqrt(tr_u)
        result.update(pclass=pclass
                     ,K=K
                     ,h_d = h_u
                     ,L_d=L_u
                     ,tr_d = tr_u
                     ,th_d=th_u
                     ,q_d=q_u
                     ,Y_d=Y_u
                     ,U_d=U_u
                     ,lmda=0
                     )
    else: #capital constraint binds; fetch constrained result
        if (pclass == 4):
            assert L_u > 4*B*v / (a2r2 + 4*v*w), \
            'LG: L_u should be greater than max unconstrained L'
        msg = "cultivator: for pclass {}, capital constraint binds".format(pclass)
        logging.debug(msg)
        #switch to constrained result:
        result = constrained_cultivator(agent, B, K, v, w, rho, params)
    return result

def constrained_cultivator(agent, B, K, v, w, rho, params):
    """Return dict of agent attributes,
    including pclass, factor demands, and welfare.
    """
    assert B==agent.bbar + v*agent.hbar - K
    B1, B2, B3 = b123(agent, K, v, w, rho, params)
    assert (B >= B1)
    if B1 <= B <= B2:  # Producer is self_cultivator 
        agentinfo = self_cultivator(agent, B, K, v, w, rho, params) 
    elif B2 < B < B3:    # Producer is small capitalist 
        agentinfo = small_capitalist(agent, B, K, v, w, rho, params)
    elif (B >= B3):                # Producer is large capitalist
        agentinfo = large_capitalist(agent, B, K, v, w, rho, params)
    else:
        msg = "Bad B={} (vs B1={}, B2={}, B3={}".format(B,B1,B2,B3)
        raise ValueError(msg)
    return agentinfo
 

def large_capitalist(agent, B, K, v, w, rho, params):
    """Return dict of agent attributes,
    including pclass, factor demands, and welfare.

    Parameters
    ----------
    B : float 
        exogenous working capital (may be negative due to K)
    hbar : float
        Quantity of land owned
    K : float 
        Fixed set-up costs 
    rho : float 
        Premium associated with participation in modern value chain
    v : float 
        Price of land
    w : float
        Price of labor
    params : dict
        Model parameters, especially s1, s2, D, KM, KT.
        (see distribution.py for the definitions)

    Returns
    -------
    result : dict
        keys are pclass,tr_d,tw_d,th_d,L_d,h_d,q_d,Y_d,lmda,U_d
        (See the module documentation for definitions.)
    """
    assert (K==params['KT'] or K==params['KM'])
    result = dict(pclass=4, tw_d=0, th_d=0, constraints=0)
    B1, B2, B3 = b123(agent, K, v, w, rho, params)
    assert (B >= B3)
    hbar = agent.hbar
    A = agent.A #Productivity parameter on Cobb-Douglas production function
    s1 = params['s1']
    s2 = params['s2']
    D = params['D']
    #1. check that it pays LG to produce
    dar = D / (A * rho)
    aa2 = A*A
    rr2 = rho*rho
    a2r2 = aa2*rr2
    dd2  = D*D
    vv2 = v*v
    ar4vw01 = a2r2 - 4*v*w
    s12 = s1/(2*s2)
    # compute constrained solution
    # **tentative** constrained
    ar4vw02 = a2r2 / (4 * v * w)
    minL = (B/w) / (1 + ar4vw02)
    maxL = (math.sqrt(s1*s1+4*s2)-s1)/(2*s2)  #highest possible L value (asymptote)
    hmax = (B/v) * ar4vw02 / (1 + ar4vw02)
    ####old appraoch
    ww2 = w*w
    v04 = B*B
    v05 = 1.0/w
    v06 = 1.0/s2
    v07 = s1*w
    v08 = 1.0 / ww2
    s2s2i = 1.0 / (s2*s2)
    v10 = -(B*s2)
    v11 = 2**(-1.0/3)
    v12 = 2**(-2.0/3)
    v13 = s2*v*dd2
    v14 = s1**2*v*dd2
    v15 = v07 + v10
    v16 = 4*s1*B*v13
    v17 = v14*w
    v18 = -v17
    v19 = v15**2
    v20 = -4*ww2*a2r2
    v21 = -4*B*a2r2*v07
    v22 = s2*a2r2*v04
    v23 = v13 - a2r2*w
    v24 = 1.0/v23
    v25 = s1*B*a2r2 + v14 + 4*a2r2*w
    v26 = s1*ww2*a2r2 + B*s2**2*v*dd2 - v07*v13 + a2r2*v10*w
    v27 = -108*s2*v04*v23*v25**2*w
    v28 = v16 + v18 + v20 + v21 + v22
    v29 = 4*ww2*a2r2 + 4*B*a2r2*v07 - 4*s1*B*v13 + v17 - v22
    v30 = -432*a2r2*v04*v26**2
    v31 = (2*B*a2r2*v07 - 2*s1*B*v13 + v18 + v20 + v22)**2
    v32 = 2*v28**3
    v33 = -(v05*v06*v24*v29)/4.0
    v34 = -288*(s2*a2r2*v04 + v16 + v18 + v20 + v21)*v22*v23*w
    v35 = -36*B*v25*v26*v28
    v36 = -4*(-12 *B*v25*v26 + v28**2 + 48*v22*v23*w)**3
    v37 = v27 + v30 + v32 + v34 + v35
    v38 = np.sqrt(v36 + v37**2)
    v39 = np.sqrt((v27 + v30 + v32 + v34 + v35)**2 + v36) + v37
    v40 = v39**(1/float(3))
    v41 = v39**(-1/float(3))
    v42 = (v08*s2s2i*v19)/4.0 - (v05*v06*v24*v28)/12.0 + v33 - \
          (v05*v06*v11*v24*v40)/float(12) - (v05*v06*v12*v24*v31*v41)/6.0
    L_c_old = (v05*v06*(3*B*s2 - 3*v07- 6*s2*np.sqrt(v42)*w + 
        6*s2*np.sqrt((v08*s2s2i*v19)/float(2) + 
        (v05*v06*v24*v28)/float(12) + v33 +
        (v05*v06*v11*v24*v40)/float(12) + 
        (v05*v06*v12*v24*v31*v41)/float(6) - 
        (2*B*v05*v06*v24*v25 + v08*s2s2i*v15*v24*v29 - 
        v15**3/(s2**3*w**3))/(4*np.sqrt(v42)))*w))/float(12)
    ### new approach
    #notation
    bbw = B/w
    bw2 = bbw * bbw
    bw3 = bw2 * bbw
    bw4 = bw2 * bw2
    q0 = math.sqrt(w/v)/dar
    q2 = q0 * q0
    q4 = q2 * q2
    q6 = q4 * q2
    s12bw = s1/s2 - bbw
    #L1
    L1 = -(1./4.)*s12bw
    #L2
    L2c1 = 2*(-(s1**3*(s1 + 2*bbw*s2)**3) + q6*(-4 + 2*bbw*s1 + bw2*s2)**3 
           + 3*q4*(2*s1**2*(-8 + bbw*s1*(8 + 7*bbw*s1))
           + 8*bbw*s1*(-4 + bbw*s1*(14 + 3*bbw*s1))*s2
           + bw3*s1*(160 + 9*bbw*s1)*s2**2 - 2*bw4*(-36 + bbw*s1)*s2**3)
           + 3*q2*(2*s1**4*(-2 + bbw*s1) - bbw*s1**3*(16 + 9*bbw*s1)*s2
           - 8*bw2*s1**2*(11 + 3*bbw*s1)*s2**2
           - 2*bw3*s1*(72 + 7*bbw*s1)*s2**3 - 72*bw4*s2**4))
    #L2c2i = -4 * (s1*(s1 + 2*bbw*s2) - q2*(-4 + 2*bbw*s1 + bw2*s2))**6
    sqrtnegL2c2i = 2 * (s1*(s1 + 2*bbw*s2) - q2*(-4 + 2*bbw*s1 + bw2*s2))**3
    #print "L2c2i,L2c1,L2c2i + L2c1*L2c1", L2c2i, L2c1, L2c2i + L2c1*L2c1
    try:
        #L2c2 = math.sqrt(L2c2i + L2c1*L2c1)
        #the above can fail due to rounding error or overflow
        #but the following might still fail due to rounding error
        L2c2 = math.sqrt((L2c1+sqrtnegL2c2i)*(L2c1-sqrtnegL2c2i))
    except ValueError:
        msg = "L2c2 ValueError: v={}, w={}, rho={}: ".format(v, w, rho)
        logging.info(msg)
        print(msg)
        print(L2c1,sqrtnegL2c2i)
        assert np.allclose(L2c1,abs(sqrtnegL2c2i))
        L2c2 = 0
    L2c = L2c1 + L2c2
    L2a = 8*q2 + 8*bbw*q2*s1 + 2*s1*s1 - 2*bw2*q2*s2 - 8*bbw*s1*s2
    L2b = (-4*q2 + 2*bbw*q2*s1 - s1*s1 + bw2*q2*s2 - 2*bbw*s1*s2)**2
    pt1 = (1./4.)*s12bw**2 + L2a/(12*(q2 - s2)*s2)
    pt2 = (1./(12*(q2 - s2)*s2))*(2**(1./3.)*L2b/L2c**(1./3.) + 2**(-1./3.)*L2c**(1./3.))
    L2 = -(1./2.)*math.sqrt(pt1 + pt2)
    #L3
    L3a = L2a/(-2*(q2 - s2)*s2)
    pt3 = (-2*bbw*(4*q2 + bbw*q2*s1 + s1*s1)/((q2 - s2)*s2) - s12bw**3 + s12bw*L3a)/(8*L2)
    discL3 = 2*pt1 - pt2 + pt3
    if discL3 < 0:  #should only happen via rounding error
        print("problem with discL3 {}".format(discL3))
        #:todo: find better response to possible subtractive cancellation problem
        discL3 = 0
    L3 = (1./2.)*math.sqrt(discL3)
    #old version (kept only for reference):
    #L3 = (1./2.)*math.sqrt(2*pt1 - pt2 - (-(( 2*bbw*(4*q2 + bbw*q2*s1 + s1*s1))
    #                           /((q2 - s2)*s2)) - s12bw**3 + s12bw*L3a)/(-8*L2))
    L_c = L1 + L2 + L3  #constrained optimum
    if math.isnan(L_c) or (v<1e-8 or w<1e-8) or __debug__:
        obj2min = lambda L: \
            B + K - D*math.sqrt(1-s1*L-s2*L*L) - v*hbar - A*rho*math.sqrt(L*(B-w*L)/v)
        numL = fminbound(obj2min, 0, (math.sqrt(s1*s1+4*s2)-s1)/(2*s2))
        if numL > maxL:
            msg = "should have numL={} < maxL={}".format(numL,maxL)
            logging.warn(msg)
            assert np.allclose(numL, maxL)
            numL = maxL - 1e-12   #:note: just in case bound was not tightly enforced
    if math.isnan(L_c) or (v<1e-8 or w<1e-8):
        L_c = numL
        msg = "new L_c calculation failed; use numL"
        msg += "({numL})".format(numL=numL)
        print msg
        logging.info(msg)
    elif __debug__:
        assert np.allclose(L_c, numL), "{} vs {}+{}+{}".format(numL, L1, L2, L3) + msg
        if not math.isnan(L_c_old):
            assert np.allclose(L_c, L_c_old), "{} vs {}+{}+{}".format(L_c_old, L1, L2, L3)
            assert np.allclose(L_c_old,numL)
        else:
            msg = "old L_c calculation failed"
            logging.info(msg)
            print """v19:{v19}, v39:{v39}, v42:{v42}, B:{B}
            """.format(v19=v19,v39=v39, v42=v42,B=B)
            msg = "L_c_old={L},numL={numL}".format(L=L_c_old,numL=numL)
            print msg
            logging.info(msg)
    assert L_c > minL, "constrained solution {} should exceed {}.".format(L_c,minL)
    h_d = (B - w*L_c)/v 
    if h_d > hmax: #should only happen from rounding error
        assert np.allclose(hmax/h_d,1.0), "{} vs. {}".format(h_d, hmax)
        h_d = hmax
    #:note: set implied Lagrange multiplier
    result['lmda'] = (A*L_c**(0.5)*rho)/(2*h_d**(0.5)*v) - 1
    assert L_c>=0, 'L_c={} should be nonnegative'.format(L_c)
    ts_c = s1*L_c + s2*L_c*L_c
    if ts_c > 1: #should only happen due to rounding error
        assert np.allclose(L_c,maxL)
        msg = 'LG: L_c={0}; maxL={1}'.format(L_c,maxL)
        logging.warn(msg)
        print(msg)
        assert np.allclose(ts_c, 1.0)
        L_c = maxL - 1e-12   #:note: HACK! but o/w rounding error can cause ts>1
        ts_c = s1*L_c + s2*L_c*L_c
        assert ts_c < 1
    assert s1*L_c + s2*L_c*L_c <= 1, \
        'L_c={0} yields ts={1}'.format(L_c,s1*L_c + s2*L_c*L_c)
    assert h_d>=0, 'h_d={}'.format(h_d)
    result['L_d'] = L_d = L_c
    result['h_d'] = h_d
    result['tr_d'] = tr_d = 1 - s1*L_d - s2*L_d*L_d
    assert tr_d>=0, 'tr_d={}'.format(tr_d)
    result['q_d'] = q_d = A * math.sqrt(h_d*L_d)
    result['Y_d'] = Y_d = rho*q_d - v*(h_d - hbar) - w*L_d - K
    result['U_d'] = Y_d + D*math.sqrt(tr_d)
    return result

def small_capitalist(agent, B, K, v, w, rho, params):
    """Return dict of agent attributes,
    including pclass, factor demands, and welfare.

    Parameters
    ----------
    B : float 
        exogenous working capital (may be negative due to K)
    hbar : float
        Quantity of land owned
    K : float 
        Fixed set-up costs (equals KT or KM)
    rho : float 
        Premium associated with participation in modern value chain
    v : float 
        Price of land
    w : float
        Price of labor
    params : dict
        Model parameters, especially s1, s2, D, KM, KT.
        (see distribution.py for the definitions)

    Returns
    -------
    result : dict
        keys are pclass,tr_d,tw_d,th_d,L_d,h_d,q_d,Y_d,lmda,U_d
        (See the module documentation for definitions.)
    """
    assert (K==params['KT'] or K==params['KM'])
    #TODO: check that it pays to produce at all (rho must be high enough)
    hbar, A = agent.hbar, agent.A
    s1, s2, D = params['s1'], params['s2'], params['D']
    result = dict(pclass=3, tw_d=0, constraints=0)
    #NEW approach
    #1. FIRST: check that it is profitable to hire labor
    dar = D / (A * rho)
    aa2 = A*A
    rr2 = rho*rho
    a2r2 = aa2*rr2
    ar4vw02 = a2r2 / (4 * v * w)

    #original formulation
    R_c = (D**2*v*(-2*s2*D**2*v + A**2*rho**2*(-2*B*s2 + w - s1*w) + 
        2*np.sqrt(s2**2*(A**2*B*rho**2 + D**2*v)**2 + 
        A**2*(-1 + s1)*s2*rho**2*(A**2*B*rho**2 + D**2*v)*w + 
        A**4*((-1 + s1)**2 + 3*s2)*rho**4*w**2)))/(3*A**4*rho**4*w**2)
    Ld_c = (s2*D**2*v + A**2*rho**2*(B*s2 + w - s1*w) - 
        np.sqrt(s2**2*(A**2*B*rho**2 + D**2*v)**2 + 
        A**2*(-1 + s1)*s2*rho**2*(A**2*B*rho**2 + D**2*v)*w + 
        A**4*((-1 + s1)**2 + 3*s2)*rho**4*w**2))/(3*A**2*s2*rho**2*w)

    #preferred formulation
    x = (B + dar*dar*v) / w
    L_d = (1-s1+s2*x - math.sqrt((1-s1-x*s2)**2+3*s2+(1-s1)*x*s2))/(3*s2)
    tr_d = (1-s1-2*s2*L_d)*dar*dar*v/w


    #yet another (Mma based) formulation
    L201506 = (D**2*s2*v
             + A**2*rho**2*(B*s2 + w - s1*w)
             - math.sqrt(3*A**2*rho**2*s2*w* (D**2*(-1 + s1)*v + A**2*rho**2*(B*(-1 + s1) + w))
                    + (D**2*s2*v + A**2*rho**2*(B*s2 + w - s1*w))**2)
            ) / (3.*A**2*rho**2*s2*w)
    tr201506 = (D**2*v*(
                        -2*D**2*s2*v
                        + A**2*rho**2*(-2*B*s2 + w - s1*w)
                        + 2*math.sqrt(3*A**2*rho**2*s2*w
                                      * (D**2*(-1 + s1)*v + A**2*rho**2*(B*(-1 + s1) + w))
                                      + (D**2*s2*v + A**2*rho**2*(B*s2 + w - s1*w))**2
                                     )
                       )
               ) / (3.*A**4*rho**4*w**2)

    #these are complicated enough to check various expressions against each other:
    assert np.allclose(L201506, Ld_c)
    assert np.allclose(L_d, Ld_c)
    assert np.allclose(tr201506, R_c)
    assert np.allclose(tr_d, R_c)

    h_d = (B - w*L_d) / v
    assert B < ((1 - s1)*w)/(2*s2) \
        + (A*A*(1-s1)*(1-s1)*rho*rho)/(16*s2*v) - (3*v*w*w)/(A*A*s2*rho*rho) \
        + (rho*rho*A*A*(1-tr_d))/(4*v), "see solutions"  #:note: *unconstrained* tr_d here
    #TODO: Simplify lmda?
    lmda = np.sqrt(v*w**2*(-2*s2*D**2*v + A**2*rho**2*(-2*B*s2 + w - s1*w) + 
       2*np.sqrt(s2**2*(A**2*B*rho**2 + D**2*v)**2 + 
       A**2*(-1 + s1)*s2*rho**2*(A**2*B*rho**2 + D**2*v)*w + 
       A**4*((-1 + s1)**2 + 3*s2)*rho**4*w**2)))/(2*np.sqrt(3)*v*w**2) - 1
    if L_d < 0:
        msg = "L={L}".format(L=L_d)
        logging.warn(msg)
        print "LG:", msg
        L_d = 0
    if tr_d < 0:
        msg = "tr_d={tr_d}".format(tr_d=tr_d)
        logging.warn(msg)
        print "LG:", msg
        tr_d = 0
    if tr_d > 1:
        result['constraints'] |= constraints['R1']
        logging.warn('LG: time constraint limits tr_d to 1.')
        print('LG: time constraint limits tr_d to 1.')
        tr_d = 1
    if h_d < 0:
        msg = "h={h}".format(h=h_d)
        logging.warn(msg)
        print "LG:", msg
        h_d = 0
    ts_d =  s1*L_d + s2*L_d*L_d 
    th_d = 1 - tr_d - ts_d
    if th_d < 0:
        msg = "th_d={th_d}".format(th_d=th_d)
        logging.warn(msg)
        print "LG:", msg
        th_d = 0  
    result['L_d'] = L_d 
    assert L_d>=0, 'L_d={}'.format(L_d)
    result['tr_d'] = tr_d   
    assert 1>=tr_d>=0, 'tr_d={}'.format(tr_d)
    result['h_d'] = h_d  
    assert h_d>=0, 'h_d={}'.format(h_d)
    result['th_d'] = th_d
    assert 1>=th_d>=0, 'th_d={}'.format(th_d)
    result['ts_d'] = ts_d
    assert 1>=ts_d>=0, 'ts_d={}'.format(th_d)
    result['q_d'] = q_d = A * math.sqrt(h_d*(th_d+L_d))
    result['Y_d'] = Y_d = rho*q_d - v*(h_d-hbar) - w*L_d - K
    result['U_d'] = Y_d + D*math.sqrt(tr_d)
    result['lmda'] = lmda
    return result

def self_cultivator(agent, B, K, v, w, rho, params):
    """Return dict of agent attributes,
    including pclass, factor demands, and welfare.

    Parameters
    ----------
    hbar : float
        Quantity of land owned
    B : float 
        exogenous working capital (may be negative due to K)
    K : float 
        Fixed set-up costs (equals KT or KM)
    rho : float 
        Premium associated with participation in modern value chain
    v : float 
        Price of land
    w : float
        Price of labor
    params : dict
        Model parameters, especially s1, s2, D, KM, KT.
        Here we need A and D.
        (See distribution.py for the definitions)

    Returns
    -------
    result : dict
        keys are pclass,tr_d,tw_d,th_d,L_d,h_d,q_d,Y_d,lmda,U_d
        (See the module documentation for definitions.)
    """
    assert (K==params['KT'] or K==params['KM'])
    assert B>0, "if B<=0, cannot cultivate"
    assert v>0, "if v<=0, problem ill defined; solutions nonsense"
    hbar = agent.hbar
    A = agent.A #Productivity parameter on Cobb-Douglas production function
    D = params['D']
    dar = D / (rho * A)
    result = dict(pclass=2, tw_d=0, L_d=0, constraints=0)
    #CONSTRAINED and UNCONSTRAINED SOLUTIONS
    h_c = B/v
    h_u = (A*A*rho*rho)/(4*v*v) - dar*dar #unconstrained
    assert (h_u > 0)
    if h_u <= 0: #does not pay to produce, delegate to PAL
        result.update(pure_laborer(agent, v, w, params)) #sets pclass=0
        result['U_d'] -= K
        result['lmda'] = 0
        #result['constraints'] |= constraints['R1']
        logging.warn('self_cultivator: delegate to pal')
    # note: th_d->0 as h->0 (in both cases)
    assert (h_c < h_u)  #usual case: working capital constraint binds
    assert 0 < B < (rho*rho*A*A/(4*v) - dar*dar*v), \
        "violated upper bound on B for constraint to bind"
    result['constraints'] |= constraints['B']
    h_d = h_c
    th_d = h_c/(dar*dar+h_c) #constrained value
    assert np.allclose(th_d, A*A*h_d*rho*rho/(D*D+A*A*h_d*rho*rho)) #0 < th_d < 1
    lmda = A*rho/(2 *v*math.sqrt(dar*dar+B/v)) - 1
    lmda02 = (A**2*rho**2)/(2*v**(0.5)*(A**2*B*rho**2 + v*D**2)**(0.5)) - 1
    lmda03 = 0.5*A*rho*math.sqrt(th_d/(v*B)) - 1
    if not np.allclose(lmda, lmda02, lmda03):
        print "sign problem? {} vs {} vs {}".format(lmda,lmda02, lmda03)
        print A, rho, v, B
        raise ValueError()
    #record SC solutions
    if result['pclass'] == 2: #o.w. delegated to PAL
        result['h_d'] = h_d
        result['th_d'] = th_d
        result['tr_d'] = tr_d = 1 - th_d   #1 if h=0
        result['q_d'] = q_d = A * math.sqrt(h_d*th_d)  #0 if h=0
        result['Y_d'] = Y_d = rho*q_d + v*(hbar - h_d) - K
        if h_c > h_u:
            assert np.allclose(Y_d, th_d*A*A*rho*rho/(4*v) + v*hbar - K)
        result['U_d'] = Y_d + D*math.sqrt(tr_d)
        result['lmda'] = lmda
    else:
        assert result['pclass'] == 0, "only delegation is to PAL"
    return result

def laborer_cultivator(agent, B, K, v, w, rho, params):
    """Return dict of agent attributes,
    including pclass, factor demands, and welfare.

    Parameters
    ----------
    hbar : float
        Quantity of land owned
    B : float 
        exogenous working capital (may be negative due to K)
    K : float 
        Fixed set-up costs 
    rho : float 
        Premium associated with participation in modern value chain
    v : float 
        Price of land
    w : float
        Price of labor
    params : dict
        Model parameters, especially s1, s2, D, KM, KT.
        Here we need A and D.
        (See distribution.py for the definitions)

    Returns
    -------
    result : dict, laborer_cultivator's attributes,
        keys are pclass,tr_d,tw_d,th_d,L_d,h_d,q_d,Y_d,lmda,U_d
        (See the module documentation for definitions.)
"""
    assert (K==params['KT'] or K==params['KM'])
    B1, B2, B3 = b123(agent, K, v, w, rho, params)
    #B < B1 = w(1-tr_d) here
    assert B<B1, "caller should enforce B<B1"
    hbar = agent.hbar
    A = agent.A #Productivity parameter on Cobb-Douglas production function
    D = params['D']
    dar = D / (rho * A)
    result = dict(pclass=1, L_d=0, constraints=0)
    #if it pays to produce then capital constraint binds, otherwise ...
    if (rho * A <= 2 * math.sqrt(v*w)) or (B+B1<=0): #doesn't pay to cultivate (see below)
        result.update(pure_laborer(agent, v, w, params))
        #fixed costs already incurred
        result['U_d'] -= K
        result['Y_d'] -= K
        result['lmda'] = 0   #:note: perhaps None wd be more informative?
        if (rho * A <= 2 * math.sqrt(v*w)): #this happens a lot!
            result['constraints'] |= constraints['rA']
            msg = "LC: rho*A too low to cultivate; delegate to PL."
            logging.debug(msg)
        if (B+B1<=0): #happens *very* often (at least out of eq)
            result['constraints'] |= constraints['ell0']
            msg = "LC: can't afford land to cultivate; delegate to PL."
            logging.debug(msg)
    else:  #:note: if it pays to be an LC, working capital constraint binds
        tr_d, tw_d, th_d, h_d = 1 - B1/w, (B1-B)/(2*w),  (B1+B)/(2*w), (B1+B)/(2*v)
        #as -B1 < B < B1, all vars are properly bounded
        assert np.allclose([tr_d,tw_d,th_d,h_d]
                          ,[(v/w)*dar*dar,(1-tr_d-B/w)/2,(1-tr_d+B/w)/2,(w/v)*th_d]
                          )
        result['tr_d'] = tr_d
        result['tw_d'] = tw_d
        result['th_d'] = th_d
        result['h_d'] = h_d
        assert np.allclose(v*h_d, w*th_d) #equal expenditure on each factor
        result['q_d'] = q_d = A * math.sqrt(h_d*th_d)
        result['Y_d'] = Y_d = rho*q_d + w*tw_d + v*(hbar-h_d) - K
        assert np.allclose(q_d, A*(1 - tr_d + B/w)*math.sqrt(w/v)/2.,
            A*(B1 + B)/math.sqrt(4*w*v))
        assert np.allclose(Y_d, 0.5*A*rho*(B+w*(1-tr_d))/math.sqrt(v*w) + v*hbar - B- K)
        result['U_d'] = Y_d + D*math.sqrt(tr_d)
        result['lmda'] = A*rho/(2*math.sqrt(v*w))-1
    return result

def pure_laborer(agent, v, w, params):
    """Return dict of agent attributes,
    including pclass, factor demands, and welfare.

    Parameters
    ----------
    agent : Producer or AgentInfo
      we need access to the hbar attribute
      (quantity of land owned)
    v : float
      price of land
    w : float
      price of labor
    params : dict
      model parameters;
      we just need access to D.
      (see baseline.ini for the definitions)

    Returns
    -------
    result : dict, pure_laborer's attributes,
        keys are pclass,tr_d,tw_d,th_d,L_d,h_d,q_d,Y_d,lmda,U_d
        (See the module documentation for definitions.)
    """
    assert (v > 0 and w > 0),\
        "negative factor prices: v={}, w={}".format(v,w)
    result = dict(pclass=0
                 ,K=0
                 ,th_d=0
                 ,ts_d=0
                 ,L_d=0
                 ,h_d=0
                 ,q_d=0
                 ,constraints=0
                 )
    D = params['D'] #Parameter on sub-utility (of leisure) function
    d2w = D / (2.0 * w)
    #optimum to pure laborer's objective, subject to tr_d<=1
    if (d2w < 1.0):
        tr_d = d2w * d2w
    else:
        tr_d = 1.0
        result['constraints'] |= constraints['R1']
    result['tr_d'] = tr_d
    result['tw_d'] = tw_d = 1 - tr_d    #PL has only two uses of time
    result['q_d'] = 0.0
    result['Y_d'] = Y_d = w*tw_d + v*agent.hbar #no production by PL
    Utr = D*d2w if (tr_d<1) else Y_d + D #faster than sqrt(tr_d)
    result['U_d'] = Y_d + Utr
    result['lmda'] = None        #capital constraint irrelevant to PL
    return result


# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 autoindent
