"""File: distribution.py

Provides classes used in:
Modern Value Chains and the Organization of Agrarian Production

Classes
-------
AgentInfo : a namedtuple holding core agent attributes
MktInfo : a namedtuple holding a basic market summary
Producer : a class for the specification of agent attributes and behavior
Procurer : a class for the specification of firm attributes and behavior
World : a class for the creation, coordination, and documentation of agents
GUI : a class for GUI set-up (only if desired; not needed to run model)

Dependencies
------------
Python 2.7 : https://www.python.org/
numpy : http://numpy.scipy.org/
scipy : http://www.scipy.org/
gridworld.py : http://code.google.com/p/econpy/source/browse/trunk/abm/
    (in the baseline, this is just for some GUI conveniences;
    in extensions, provides a topology) (place in same folder)
choose.py : included module (place in same folder)
parameters.py : included module (place in same folder)

:note: documented sided effects are changes in program state;
  usually we won't bother to document IO (esp. user msgs)
"""
from __future__ import division
#standard library imports:
import itertools, logging, math, sys
from collections import Counter, deque, namedtuple, defaultdict
#commonly used scientific libraries:
import numpy as np
from scipy.optimize import fminbound, fsolve
from scipy.optimize import minpack #only for trapping minpack.error
#other imports:
import gridworld
import choose
from utilities import gini, incremental_shares_pareto
from utilities import shrink_bracket, simple_bracket_search

__version__ = "2.1.9"

#REQUIRE Python 2.7.x
if sys.version_info[0] != 2 or sys.version_info[1] < 7:
    print("This script requires Python version 2.7")
    sys.exit(1)

############################  global constants   ##############################
###############  (these are largely for reporting convenience) ################

###############  macro-level variables (see `log_macrodata` for definitions)
macro_datanames = tuple("""
delta_pct n_agents KF PF
KM cdistscale theta
v w rho
H hsumPL hsumLC hsumSC hsumSM hsumLG
Qtotal Qav QMtotal QMav
Ytotal Yav YMtotal YMav
Utotal Uav UMtotal UMav Ulandless 
p_qmodern p_ymodern Pi
n_modern n_poor n_contract
n_class0 n_class1 n_class2 n_class3 n_class4
n_sector0 n_sector1 n_sector2
n_frenege n_prenege n_accept n_honor
p_modern p_poor p_contract p_qmodern
hd_total twd_total thd_total Ld_total tsd_total
gini_y gini_u
""".split())

###############  agent-level variables
micro_datanames = tuple("""
delta_pct pclass sector
hbar h_d h th th_d tw tw_d L_d L tr tr_d q
is_poor U_d U Y_d Y contract
""".split())

################ contract-related bit flags
CONTRACTS = dict(
    offer=(1<<0),
    accept=(1<<1),
    fhonor=(1<<2),
    frenege=(1<<3),
    prenege=(1<<4),
    )



##############################  classes   #####################################

#provide a container for immutable agent attributes:
AgentInfo = namedtuple('AgentInfo', 'hbar bbar A')

#provide a container for recording basic mariket info
MktInfo = namedtuple('MktInfo', 'delta_pct v w rho growers')

class Producer(gridworld.Agent):
    """Provides a producing agent,
    which may be landless or landed, laborer or capitalist.
    Inheriting from `gridworld.Agent` is not important
    for the baseline model but provides GUI convenience.
    """
    _A = None  #set once in `initialize` (and nowhere else)
    #the next two attributes are reset each delta by `set_landholdings`
    #(however for the duration of that simulation, they are final)
    _hbar = 0.0 #quantity of land owned
    _bbar = 0.0 #land based credit

    def initialize(self):
        """Return None; initalize the producer attributes.
        :side-effects: set hbar, bar; set A (as final value);
            change state of prng (if `productionshocks` is True).
        :note: `initialize` is called by `gridworld.Agent.__init__`
        """
        #the model parameters are an agent property (& not agent specific)
        params = self.params
        A = params['A']  #technology without production shocks
        #extension: allow production shocks
        if params['productionshocks']: #False in baseline
            prng = params['prng']  #see parameters.py
            A *= prng.lognormvariate(0,0.01)
        self._A = A  #from now on, access via read-only property `A`
        #arbitrary display conveniences (GUI only)
        if self.world.has_gui:
            self.display(fillcolor='white', shape='square')

    def calc_info(self, v, w, rho, has_contract):
        """Return dict: agent's demands and supplies,
        pclass, sector, factor demands, and welfare.
        :calls: `choose`  (which is called nowhere else)
        :side-effects: none (does NOT set prdr info!)

        Parameters
        ----------
        v (float) : rent of land (positive float)
        w (float) : wage of labor (positive float)
        rho (float) : price premium of modern sector (positive float);
            must be in [1,PF]
        has_contract (bool) : True if agent is offered a contract

        Note
        -------
        The keys of the returned dict are:
        sector,pclass,tr_d,tw_d,th_d,L_d,h_d,q_d,Y_d,lmda,U_d
        (See the `choose.py` module documentation for definitions.)
        """
        #enforce pre-conditions
        assert (w>0 and v>0),\
            "bad factor prices: v={}, w={}".format(v, w)
        assert (has_contract in (False, True)),\
            "uknown contract state {}".format (has_contract)
        params = self.params  #the model params (not agent specific)
        assert (rho>=1 and rho <= params['PF']), "bad rho: {}".format(rho)
        #
        KM, KT = params['KM'], params['KT']  #fixed costs of sector participation
        #K=0 for pure laborer; see below
        #get the info (dict) for each possible choice
        agent = self.baseInfo #read-only, just to make a point (and add safety)
        lab = choose.pure_laborer(agent, v, w, params)
        #traditional -> K=KT,rho=1.0
        trad = choose.cultivate(agent, params['KT'], v, w, 1.0, params)
        # Growers under contract have the full range of options:
        if has_contract and rho>1:
            #modern -> K=KM,rho=rho
            mod = choose.cultivate(agent, params['KM'], v, w, rho, params)
            sectors = (lab, trad, mod)
        else:     # Growers not under contract; 'mod' is not available
            sectors = (lab, trad)
        #choose sector to maximize utility, with implied fixed costs
        info = max(sectors, key=lambda s: s['U_d'])
        info['sector'] = sectors.index(info)
        assert "contract" not in info
        info['contract'] = 0
        if has_contract:
            info['contract'] |= CONTRACTS['offer']
        if info['sector'] == 0:
            assert info['K'] == 0
        elif info['sector'] == 1:
            assert info['K'] == KT
        elif info['sector'] == 2:
            assert info['K'] == KM
            assert has_contract
            info['contract'] |= CONTRACTS['accept']
        else:
            raise ValueError("unknown sector".format(info['sector']))
        info['vwrho'] = (v, w, rho)
        info['hbar'] = agent.hbar  #only needed for microdata logging
        #next we record an arbitrary value of lmda for those who choose `lab`
        if info is lab:  # lmda does not exist for laborers; arbitrary replacement
            info['lmda'] = 0
        return info


    def qYU(self, info):
        """Return 3-tuple of float,
        the producer's output, income, and utility.
        :note: info must include ex post values 
          (which allow for short-side constraints),
          provided by `updated_producer_infos`
        """
        hbar = self.hbar
        assert hbar == info['hbar'] #id's the producer
        #compute q
        h = info['h']
        L = info['L']
        th = info['th']
        #:note: A may be individualized (but not in baseline)
        q = self.A * math.sqrt(h * (th + L))
        #compute Y
        v, w, rho = info['vwrho']
        yv = v * (hbar - info['h'])
        yw = w * (info['tw'] - info['L'])
        if bool(info['contract'] & CONTRACTS['accept']):
            assert bool(info['contract'] & CONTRACTS['offer'])
            price = rho
        else:
            price = 1.0
        yq = price * q
        Y = yq + yv + yw - info['K']  #see `calc_info`
        D = self.params['D']
        U = Y + D * math.sqrt(info['tr'])
        return (q, Y, U)

    ### read-only properties ###

    @property
    def params(self):
        return self.world.params

    @property
    def A(self):
        #`_A` is set in `initialize`
        assert (self._A == self._baseInfo.A)
        return self._baseInfo.A

    @property
    def hbar(self):
        return self._baseInfo.hbar

    @property
    def bbar(self):
        return self._baseInfo.bbar

    #minor safety features

    @property
    def info(self):
        raise NotImplementedError

    @property
    def baseInfo(self):
        """Return AgentInfo, the core characteristics of this agent
        given the land allocation, as a namedtuple.
        :called by: `choose`
        :note: this provides all the agent info needed by `choose`
        """ 
        return self._baseInfo

    @baseInfo.setter
    def baseInfo(self, ntpl):
        assert (ntpl.hbar >= 0)
        assert (ntpl.bbar >= 0)
        assert (ntpl.A > 0)
        self._baseInfo = ntpl

    # GUI stuff

    def redraw(self, info):
        """Return None.  Reset agent display attributes.
        Side effects: GUI only.
        """
        if not self.world.has_gui:
            return
        sector = info['sector']
        contract = info['contract']
        fillcolor = ('#FFFFFF','#EEEEFF','#0000FF')[sector] #lab, trd, mod = 0, 1, 2
        if sector==2 and (CONTRACTS['frenege'] & contract):
            fillcolor = '#FF0000'   #color freneged (color choice? chk)
        size = (self.hbar + 0.01)**(1/6.0)  #display size reflects land owndership
        self.display(fillcolor=fillcolor, shape='square', shapesize=(size, size))




class Procurer(gridworld.Agent):
    """Provides a modern-sector firm that can contract with growers.
    Inheriting from `gridworld.Agent` is not important
    for the baseline model but provides GUI convenience.
    """

    def initialize(self):
        """Return None. Set miscellaneous initial values.  
        :note: this method is called by gridworld.Agent.__init__
        """
        self.reset()
        #for GUI user convenience, make firm gray until it chooses a location
        if self.world.has_gui:
            self.display(fillcolor='gray', shape='circle', shapesize=(1,1))

    def reset(self):
        """Return None.
        :side-effects: Reset growers attributes to initial state.
        """
        self._growers = None
        self._honored = None
        self._freneged = None
        self.rho = 1
        self.Pi = 0

    def chooseLocation(self, v, w):
        """Return None.  Set position to profit maximizing location.
        :side-effects: prng state changes
          (via`random_locations`) if `cdistscale` > 0
          and via `rho_epi` *if* `extrasearch` is True.
        :note: irrelevant to baseline model

        Parameters
        ----------
        v (float) : rent of land
        w (float) : wage of labor
        """
        world = self.world
        params = self.params
        if (params['cdistscale'] == 0):
            msg = """cdistscale==0;
            distance doesn't matter so location doesn't matter;
            we will leave firm at its initial location."""
            world.logger.info(msg)
            return self.position
        #   ^^^^^^
        print("It can take a while for the procurer to choose a location ...")
        #random possible firm locations
        n_locations = params['n_locations']
        #get a random set of unoccupied firm locations
        locations = world.random_locations(n_locations, exclude=True, prng=params['prng'])
        loc2rho = dict()        #map of locations to best rho (given v,w)
        loc2epi = dict()        #map of locations to max profits (given v,w)
        for loc in locations:   #find best rho for each location
            reg = self.rho_epi(v, w, loc=loc)
            loc2rho[loc] = reg['rho']
            loc2epi[loc] = reg['epi']
        print("Location chosen.")
        return max(loc2epi, key=loc2epi.get) #choose best position

    def profitvwg(self, rho):
        """Return (float,float,float,list[Producer],dict[Producer,dict]):
        profit, v, w, growers, infodicts.
        Finds the best set of growers for this `rho`,
        and also the associated factor prices.
        rho (positive float) : the relative price offered to growers
        :called by: `findeq_modern` (repeatedly)
        :todo: change growers to tuple instead of list?
        :note: the challenge here is that the selected set of growers
          determines the factor market outcomes, thereby affecting
          profitability
        :note: we don't bother to filter out the nonpartipating growers
        """
        world = self.world
        #first look at the outcome with no modern sector
        if (rho <= 1 or rho >= self.params['PF']):
            v, w = world.clear_factor_markets_traditional()
            best = (0, v, w, ())  #best-so-far is 0 profit, no growers
            return best   #EARLY RETURN
        #   ^^^^^^
        best = (-1, None, None, None) #dummy initialization
        growers = []
        growersHist = [] #list of lists (past grower selections)
        ct = 0
        #look for a (best-profit, grower) configuration
        while (ct<100): #2 or 3 tries is typical, but this allows many
            if (growers in growersHist):  #stopping condition (complete cycle)
                msg = "grower set determined in {} tries".format(ct)
                logging.info(msg)
                break
            #it's a new grower list -> try contracting with these growers
            #(the issue is: factor prices will change)
            growersHist.append(growers)
            v,w = world.clear_factor_markets(rho, growers)
            #get profits when *all* `growers` contracts fhonored (!)
            # along with the preferred growers at v,w, rho (out of all producers)
            #note: do NOT use `growers_and_profit` instead!
            gp = self.profitFhonor(v, w, rho, growers)
            profit = gp['profit'] #profit when *all* `growers` contracts fhonored
            if profit > best[0]: #then update best
                best = (profit, v, w, growers)
            #NOTE: `growers` (above ^) induced this (v,w)
            #   but perhaps they aren't the best set for this v,w,rho!
            #(-> new `growers` (below v) may not be highest profit (once v,w change))
            growers = gp['growers']
            ct += 1
        msg = """At rho={}, we find the best profit={} after {} tries."""
        msg = msg.format(rho, best[0], ct)
        logging.info(msg)
        print(msg)
        return best

    def profitFhonor(self, v, w, rho, growers, loc=None):
        """Return dict, including the when **all** `growers` contracts fhonored:
        'profit' ->the procurer's profit (at `loc`) when *all* contracts fhonored,
        'growers' ->the profitable growers out of *all producers*,
        'infodict' ->prdr infos at (v,w,rho) if all prdrs given contracts
                     (not currently used)

        Parameters
        ----------
        v (float) : rent of land (positive float)
        w (float) : wage of labor (positive float)
        rho (float) : price premium of modern sector (positive float)
        growers (list[Producer]) : producers that "have contracts"
        loc (int,int) : proposed location of firm; default is `self.position` 

        The profit reflects honoring **all** `growers` contracts,
            so the returned level of profits can be negative.
            To reiterate, in this subroutine the procurer honors all contracts
            (buys all offer from `growers`); contrast with `growers_and_profit`.
        :note: Profits are "expected" only in the sense of expecting
          `v` and `w` unchanged.
        :note: it is really sensible to call this function
          ONLY when v, w, are market clearing values for rho, growers
        :note: previously named `profit02`
        :called-by: `profitvwg`
        :side-effects: none
        """
        #enforce precondition
        assert (v>0 and w>0),\
            "factor prices (v={},w={}) shd be positive".format(v,w)
        #
        params = self.params
        PF, KF = params['PF'], params['KF']
        if (rho <= 1 or rho > PF):
            return dict(profit=0, growers=(), infodict=None)  #EARLY RETURN
        #   ^^^^^^
        if loc is None: #:note: location does not matter in the paper simulations
            loc = self.position
        world = self._world
        cdist, producers = world.cdist, world.producers
        if cdist > 0: #cdist=0 in baseline
            distances = self._world.producer_distances(loc)
        #get producer infos supposing ALL producers were to have modern option:
        infodict = world.producer_infos(v=v, w=w, rho=rho, growers=producers)
        desirables = list()  #the desirable (profitable) producers
        total_profit = 0
        for prdr in producers: #iterate over producer list (deterministic order)
            info = infodict[prdr]
            contract = info['contract']
            assert (contract & CONTRACTS['offer'])  #every prdr has an offer
            #determine the actual profit from the existing contracts
            #and also the producers that would prove profitable growers
            if info['sector'] == 2: #if prdr wd accept an offer (ow no profit)
                assert (contract & CONTRACTS['accept'])
                #profit from this one producer (KF=0 in baseline)
                profit01 = (PF-rho) * info['q_d'] - KF
                if (cdist > 0): #no distance costs in baseline
                    profit01 -= cdist * distances[prdr]
                #determine if the procurer would like to contract with prdr
                if (profit01 > 0):
                    desirables.append(prdr) #a desired grower at v,w,rho
                #grower contracts are fhonored (even when unprofitable)
                if prdr in growers: #a grower (producer with a contract)
                    total_profit += profit01  #all contracts honored
        #check postcondition
        assert all(2==infodict[prdr]['sector'] for prdr in desirables)
        #
        return dict(profit=total_profit, growers=desirables, infodict=infodict)

    def rho_epi(self, v, w, loc=None):
        """Return dict,
        'rho'->the best rho,
        'epi' -> expected profit,
        'growers'->growers at (v,w,rho),
        Given the specified location (default: current position),
        where 'best' means greatest expected profit,
        Note: when we consider firm location & positive distance costs,
        this is called repeatedly with *ex ante* `v` and `w`
        (i.e., prior to modern sector) to determine optimal location.
        Side effect: prng state, if extrasearch used;
          see baseline.ini for details.

        Parameters
        ----------
        v (float) : rent of land, should be positive
        w (float) : wage of labor, should be positive
        loc (int,int) : procurer location for computation; optional
        """
        assert (v > 0 and w > 0), "bad factor prices ({},{})".format(v,w)
        if loc is None:
            loc = self.position
        params = self.params
        PF = params['PF']      #world price
        #get expected profit function as local var -> avoid repeated attribute access
        profite = self.profite
        f2min = lambda p: -profite(v=v, w=w, rho=p, loc=loc)
        #offering rho<1 contracts pointless; rho>PF contracts counterproductive
        results = fminbound(f2min, 1.0, PF, xtol=1e-08,full_output=True,disp=0)
        rho = results[0]  #best premium price for procurer
        EPi = -results[1]  #expected profit
        if __debug__:
            msg = "Best rho={} given v={} and w={} (using fminbound)."
            print msg.format(rho,v,w)

        if rho > PF:
            msg = "Bad fminbound result: rho={}".format(rho)
            raise ValueError(msg)

        #for more deterministic results, extrasearch turned off in the baseline
        extrasearch = self.params['extrasearch']
        if extrasearch: #search nearby for better values
            msg = "rho={} -> profits={}; trying to improve on this..."
            world.logger.info(msg.format(rho,EPi))
            rvals = [rho]
            pvals = [EPi]
            prng = params['prng']
            for _ in range(25):
                #repeatedly try a random value nearby
                rtemp = 1 + prng.lognormvariate(math.log(rho-1), 0.5)
                ptemp = profite(v, w, rtemp, loc)
                if ptemp > EPi:
                    EPi = ptemp
                    rho = rtemp
                    msg = "Improved by random search\n\t{}->{}".format(rtemp,ptemp)
                    world.logger.info(msg)
                #try to do even better with a fitted value
                if ptemp > 0: #collect values for a polynomial fit effort
                    rvals.append(rtemp)
                    pvals.append(ptemp)
                if len(rvals) > 2:
                    a,b,c=np.polyfit(rvals,pvals,2)
                    if (a < 0):
                        rtemp = -b/(2*a)
                        ptemp = profite(v, w, rtemp, loc)
                        if ptemp > EPi:
                            EPi = ptemp
                            rho = rtemp
                            msg = "Improved by structured search\n\t{}->{}"
                            msg = msg.format(rtemp,ptemp)
                            rvals.append(rtemp)
                            pvals.append(ptemp)
                            world.logger.info(msg)
        """
        We may need a final function call!
        Why? Because he chosen rho returned
        need not be identical to the last rho passed by fminbound.
        Therefore, we choose again (surprisingly, even w/o extrasearch, this can matter).
        """
        assert EPi == profite(v, w, rho, loc)
        gp = self.growers_and_profit(v, w, rho, loc=loc)
        assert np.allclose(EPi, gp['profit'])
        return dict(rho=rho, epi=EPi, growers=gp['growers'])

    def profite(self, v, w, rho, loc=None):
        """Return float, the procurer's maximum expected profit,
        given its location. (Expected only in the sense of expecting
        `v` and `w` unchanged after the contracts are offered,
        which will be true in the general equilibrium.
        :side-effects: none

        Parameters: see `growers_and_profit`

        :note: During setup, this may be called repeatedly 
          by `rho_epi` with *provisional* firm locations.
        """
        gp = self.growers_and_profit(v, w, rho, loc=loc)
        return gp['profit']

    def growers_and_profit(self, v, w, rho, loc=None):
        """Return dict, mapping (when *all* producers can opt in)
        'profit'->the procurer's maximum expected profit (at `loc`) &
        'growers'->the procurer's chosen (i.e., profitable) growers
        to generate this profit level.
        Profits are "expected" only in the sense of expecting
        `v` and `w` unchanged after the contracts are offered,
        which will be true in the general equilibrium.
        :note: a set of contracts is NOT imposed;
          contrast with `profitFhonor`
        :side-effects: none

        Parameters
        ----------
        v (float) : rent of land (positive float)
        w (float) : wage of labor (positive float)
        rho (float) : price premium of modern sector (positive float)
        growers (list[Producer]) : producers that "have contracts"
        loc (int,int) : proposed location of firm; default is `self.position` 

        Notes
        -----
        Called by `rho_epi` which may be called *repeatedly*
        with *provisional* firm locations.
        """
        if not (v>0 and w>0): #precondition
            raise ValueError("factor prices shd be positive")
        params = self.params
        PF, KF = params['PF'], params['KF']
        if (rho <=1 or rho >= PF):
            return dict(growers=(), profit=0)
        #   ^^^^^^
        if loc is None:
            loc = self.position
        cdist, producers = self._world.cdist, self._world.producers
        if cdist > 0:
            distances = self._world.producer_distances(loc)
        #get producer info when *all* have modern option
        infodict = self._world.producer_infos(v=v, w=w, rho=rho, growers=producers)
        growers = list()  #the desirable (profitable) growers
        total_profit = 0
        for prdr in producers: #determinate order of iteration
            info = infodict[prdr]
            contract = info['contract']
            assert (contract & CONTRACTS['offer'])  #pretended every prdr has an offer
            if info['sector'] == 2: #if prdr wd accept an offer
                assert (contract & CONTRACTS['accept'])
                #this producer wants a contract, check ex ante (KF) profitability
                #to see if it will actually get one
                revenue = (PF-rho) * info['q_d']
                cost = KF
                if (cdist > 0): #no distance costs in baseline
                    cost += cdist * distances[prdr]
                profit = revenue - cost
                if (profit > 0):
                    total_profit += profit
                    growers.append(prdr)
        return dict(profit=total_profit, growers=growers)

    def choose_growers(self, v, w, rho):
        """Return list of Producer,
        the producers who appear profitable ex ante,
        given `v`, `w`, `rho`, and the location of the firm.

        :note: One might expect contract-related fixed costs
            to be incurred here, but that does not make sense
            in the EK model, where fixed costs are covered
            by factor sales at realized factor prices.
        :side-effects: none
        """
        gp = self.growers_and_profit(v, w, rho)
        return gp['growers']


    def choose_contracts(self, infodict):
        """Return an infodict, marking the growers whose contracts are honored.
        Determines firm's quantity contracted, profit, and rejected growers.  
        :note: Call this function *after* the world clears factor markets
               for firm's chosen rho.
        :note: this function is roughly a no-op in the baseline,
               where all contracts are honored
        :side-effects: none
        """
        world = self.world
        infodict = world.updated_producer_infos(infodict) #new copy
        producers = world.producers  #for determinate iteration order
        params = self.params
        KT = params['KT']
        KM = params['KM']
        PF = params['PF'] #received price per unit of output
        world = self.world
        cdist = world.cdist   #cost per unit distance (zero in baseline case)
        #check the profitability of each contract
        if (cdist > 0):
            distances = world.producer_distances(self.position)
        #Get the *ex post* info dicts, given the specification of growers
        # (i.e., contract offers); includes contract acceptances
        producers = infodict.keys()
        #are these lists useful enough to keep? ... chk
        growers, rejecters, honored, freneged, preneged = [], [], [], [], []
        total_profit = 0
        vwrho = None
        for prdr in producers:
            info = infodict[prdr]
            if vwrho is None:
                vwrho = info['vwrho']
                v, w, rho = vwrho
            else:
                assert vwrho == info['vwrho']
            contract = info['contract']
            sector = info['sector']
            if bool(contract & CONTRACTS['offer']):
                growers.append(prdr)
            if not bool(contract & CONTRACTS['accept']):
                assert info['K'] == (0,KT)[sector]
                if bool(contract):
                    assert bool(contract & CONTRACTS['offer'])
                    rejecters.append(prdr)
            else: #contract was accepted
                assert info['K'] == KM
                assert bool(contract & CONTRACTS['offer'])
                if sector != 2: #then the grower reneged (cannot happen in baseline)
                    preneged.append(prdr)
                    info['contract'] |= CONTRACTS['prenege']
                else: #grower sector is 2, so KM and KF **already** sunk
                    #producer wants to sell, check profitability (after cdist and KF)
                    revenue = (PF-rho)* info['q']
                    cost = 0  #:note: KF already sunk
                    if (cdist > 0):
                        cost += cdist * distances[prdr]
                    profit = revenue - cost
                    if profit > 0:
                        honored.append(prdr)
                        info['contract'] |= CONTRACTS['fhonor']
                        total_profit += profit
                    else:
                        freneged.append(prdr)
                        #the next is redundant; just for readers convenience
                        info['contract'] |= CONTRACTS['frenege']
                        info['sector'] = 1  #sector demotion!
        if len(freneged)>0:
            dpct = world.delta_pct
            msg = "delta {0}: {1} contracts freneged\n"
            msg = msg.format(dpct,len(freneged))
            world.logger.info(msg)
            print(msg)
        if len(preneged)>0:
            dpct = world.delta_pct
            msg = "delta {0}: {1} contracts preneged\n"
            msg = msg.format(dpct,len(preneged))
            world.logger.info(msg)
            print(msg)
        assert len(growers) == (len(rejecters) + len(honored)
                                + len(freneged) + len(preneged)), \
            """{} growers
            but {} rejecters + {} honored + {} freneged + {} preneged
            """.format( len(growers),
            len(rejecters),len(honored),len(freneged), len(preneged))
        self.honored = honored  #property ensures it's only set once
        self.freneged = freneged  #property ensures it's only set once
        return infodict


    @property
    def freneged(self):
        """list[Producer] : the producers reneged on by procurer.
        (Not in baseline model.)
        """
        return self._freneged

    @freneged.setter
    def freneged(self, seq):
        if self._freneged is None:
            self._freneged = seq
        else:
            raise ValueError('freneged already set; set once only')

    @property
    def honored(self):
        """list[Producer] : the producers not reneged on by procurer.
        (Not in baseline model.)
        """
        return self._honored

    @honored.setter
    def honored(self, seq):
        if self._honored is None:
            self._honored = seq
        else:
            raise ValueError('honored already set; set once only')

    def profitxp(self, rho, infodict):
        """Return float, the *ex post* total profit.
        State dependence: rho, position, honored, growers.
        :todo: discard as redundant?
        :note: sunk costs reduce profitxp
        :called by: log_macrodata
        :side effects: none
        """
        params = self.params
        PF = params['PF'] #received price per unit of output
        KF = params['KF'] #fixed cost per **honored** grower
        cdist = self._world.cdist   #cost per unit distance
        #check the profitability of each contract
        if(cdist > 0):
            distances = self._world.producer_distances(self.position)
        total_profit = 0
        honored = list()
        freneged = list()
        for (prdr, info) in infodict.items():
            profit = 0
            contract = info['contract']
            if bool(contract & CONTRACTS['fhonor']):
                honored.append(prdr)
                assert not bool(contract & CONTRACTS['frenege'])
                assert prdr in self.honored
                profit += (PF - rho) * info['q']
                if cdist > 0:
                    profit -= cdist*distances[prdr]
                profit -= KF
            elif bool(contract & CONTRACTS['frenege']):
                freneged.append(prdr)
                profit -= KF   #:note:
            total_profit += profit
        if bool(honored):
            assert honored == self.honored,\
            "#honored {}, #self.honored {}".format(len(honored),len(self.honored))
        if bool(freneged):
            assert freneged == self.freneged
        return total_profit

    @property
    def params(self):
        return self.world.params
              
class World(gridworld.GridWorld):
    """Provides a simulation-governing agent,
    which creates, coordinates, and documents other agents.
    """ 
    setup_complete = False
    _params = None
    _procurer = None
    _producers = None
    _producer_distances = None
    _prng = None
    has_gui = False
    #delta_pct controls Pareto distn of land; see our paper
    delta_pct = 10 #starting value
    ticks = 0
    #rental rate (v) and wage rate (w) arbitrary starting values:
    #(fyi: you cannot index a deque with a slice)
    vw_init01 = deque(itertools.product([1.1 + i/10. for i in range(10)],
        [0.1 + i/10. for i in range(10)]),100)
    vw_init = deque(vw_init01, 100)
    rho_init = deque([1.35], 4)  #:note: arbitrary initialization

    def setup(self):
        """Return None.
        Do the initial setup of the entire world.
        Call setup_agents and set up log file."""
        self.logger.info('enter World.setup')
        self.setup_complete = False
        print "\nSetup in progress: please wait ..."
        self.set_derivedvals()
        self.setup_agents()
        self.setup_outfiles()
        self.test_setup()
        self.setup_complete = True
        print "setup complete: you can now run the model"
        self.logger.info('exit World.setup')

    def set_derivedvals(self):
        """Return None; set up derived values, which are final
        (i.e., these values will remain constant).
        :side-effects: set world attributes N0, N1, H, cdist
        :todo: use properties to enforce finality
        """
        params = self.params
        N = params['n_agents']
        self.N0 = N0 = int(round(N * params['p0'])) #number of landless
        self.N1 = N1 = N - N0                #number of landed
        self.H = N1//2 + N0 #(discrete population; match EK Fig 1 & 2 params)
        #the following are not relevant to the baseline scenario
        gridsize = self.topology.shape[0]
        cdist = 10. / gridsize  #basic scaling to size of grid
        cdist *= params['cdistscale']  #baseline is 0
        self.cdist = cdist

    def setup_agents(self):
        """Return None; create producer agents and procurer firm.
        Note that this setup does NOT include land distribution.
        (See `setup_run` for that.)
        :note: location does not matter in the baseline scenario
        :side-effects: created agents are attributes
        """
        #start out with the firm at the center
        #(if location matters, we'll later relocate the firm to a better spot)
        center = self.topology.shape[0] // 2
        self._procurer, = self.create_agents(
            AgentType=Procurer,
            number=1,
            locations=[(center,center)]
            )

        #create producers in random locations (which will remain fixed)
        #:note: locations do not matter at all in the baseline
        agents = self.create_agents(
            AgentType=Producer,
            number=self.params['n_agents']
            )


    def schedule(self):
        """Return None. Run one iteration.
        This is the core schedule of the simulation.
        """
        self.logger.debug("enter schedule")
        if not self.setup_complete:
            raise ValueError('must setup before running; please restart')
        # set up the new iteration by resetting all agents
        self.ticks += 1
        delta_pct = self.delta_pct #the land-distribution control parameter
        #setup for this delta: reset agents and distribute land
        self.setup_run(delta_pct/100.) #side effects: via `set_landholdings`
        msg = """
        ############# IMPLEMENT EK MODEL: NO MODERN SECTOR #############
        EXPERIMENT: {}
        CLEAR MARKETS WITH NO MODERN SECTOR: delta_pct={}
        """.format(self.params.get('name',"unknown"), delta_pct)
        print(msg)
        self.logger.info(msg)
        #start by determining traditional factor market clearing
        vxa, wxa = self.clear_factor_markets_traditional()
        self.add2vwinit(vxa, wxa, 1) #side effect: append to vw_init01
        #log the resulting data  (pre-modern, so: rho=1 and no growers)
        infodict = self.producer_infos(v=vxa, w=wxa, rho=1, growers=())
        infodict = self.updated_producer_infos(infodict)
        #naturally, there has not been any contracting (since no modern sector):
        assert not any(info['contract'] for info in infodict.values())
        ## RECORD DATA (with no modern sector present)
        mktinfo = MktInfo(delta_pct=delta_pct, v=vxa, w=wxa, rho=1, growers=())
        self.log_macrodata(self.oldmacro, mktinfo, infodict)
        if (0 == delta_pct % 10):  #log micro data every 10th iteration
            self.log_microdata(self.oldmicro, infodict)
        msg = """
        ############# INTRODUCE MODERN SECTOR #############
        EXPERIMENT: {}
        CLEAR MARKETS WITH MODERN SECTOR: delta_pct={}
        """.format(self.params.get('name',"unknown"), delta_pct)
        print(msg)
        self.logger.info(msg)
        # procurer chooses location
        firm = self.procurer
        if self.ticks==1: #on the first iteration, set firm.position
            firm.position = firm.chooseLocation(v=vxa, w=wxa) #ex ante!
            self._producer_distances = self.producer_distances(firm.position)
            if self.has_gui:
                firm.display(fillcolor='black', shape='circle', shapesize=(1,1))
        ###########   CLEAR MARKETS WITH MODERN SECTOR    ###########
        ########### (iterate to find rho,growers,v,w equilibrium) ###########
        eq = self.findeq_modern()
        v, w, rho, growers = eq['v'], eq['w'], eq['rho'], eq['growers']
        firm.rho = rho  #we could get rid of this; it's mostly for GUI ease chk
        """What follows is not relevant to the baseline model.
        In a model where the procurer can behave opportunistically
        (which requires KF > 0, among other things, which it is NOT in baseline),
        we would need to determine which contracts are honored.
        """
        infodict = self.producer_infos(v=v, w=w, rho=rho, growers=growers)
        infodict = firm.choose_contracts(infodict) # calls `updated_producer_infos`
        self.redraw_producers(infodict)  #just for the gui
        ## RECORD DATA (with modern sector present)
        mktinfo = MktInfo(delta_pct=delta_pct, v=v, w=w, rho=rho, growers=growers)
        #  record macro data:
        self.log_macrodata(self.newmacro, mktinfo, infodict=infodict)
        if (0 == delta_pct % 10): #  record micro data at select delta vals
            self.log_microdata(self.newmicro, infodict)
        """We close out this iteration with small side effects,
        which are just preparation for next iteration.""" 
        #increment the equality parameter for the next iteration (default: by 1)
        self.delta_pct += self.params['dpp']
        #finally, add v,w, rho to initial values for next iteration
        #(use as subsequent initial values, to speed convergence)
        self.rho_init.append(rho)
        self.add2vwinit(v, w, rho)
        if self.delta_pct > 99:
            self.stop()
        self.logger.debug("exit schedule\n")

    def reset(self):
        """Return None; reset the procurer attributes.
        """
        self.procurer.reset()

    def add2vwinit(self, v, w, rho):
        #This is *very* model specific: we know v and w shd be limited,
        #so we don't retry "bad" outcomes
        #:note: new starting values may affect outcomes, so use with caution
        if v < 3 and w < 2:  #don't retry "bad" outcomes
            if(rho==1):
                self.vw_init01.append((v,w))
            else:
                self.vw_init.append((v,w))
        
    def bracket_maxprofit(self):
        """Return 6-tuple of float (the bracket).
        """
        PF = self.params['PF']
        f2min = lambda p: -self.procurer.profitvwg(rho=p)[0]
        rho_best = np.mean(self.rho_init)  #somewhat arbitrary starting value
        profit_best = self.procurer.profitvwg(rho=rho_best)[0]
        #protect against the best being nevertheless bad
        if profit_best <= 0: #in this case we will try a line search
            step = (PF - 1) / 10
            rhos = [1 + step * (i+1) for i in range(9)]
            pvwgis = [self.procurer.profitvwg(rho) for rho in rhos]
            profits = list(pvwgi[0]  for pvwgi in pvwgis)
            assert all((profit >= 0) for profit in profits)
            idx = profits.index(max(profits))
            profit_best = profits[idx]
            rho_best = rhos[idx]
        if profit_best <= 0: #it does not pay the procurer to operate
            bracket = None
        else:
            bracket = shrink_bracket(f2min, 1, rho_best, PF, 0, -profit_best, 0)
        return bracket

    def findeq_modern(self):
        """Return tuple: (v, w, rho, growers) characterizing equilibrium:
        (rho, growers) is the firm's best response to (v,w),
        which *correspondingly* is a factor market equilibrium.
        :note: offering rho<1 contracts pointless; rho>PF contracts counterproductive,
          so we can bracket the best rho within that interval
        :note: currently prints lots of notes as user info (TODO: quiet mode)
        :calls: profitvwg
        :side-effects: calls self.reset()
        """
        self.reset()  #:note: just resets procurer
        bracket = self.bracket_maxprofit()
        if bracket is None: #modern sector is not profitable
            v, w = self.clear_factor_markets_traditional()
            return dict(profit=0, v=v, w=w, rho=1, growers=())
        #   ^^^^^^
        else:
            (xa, xm, xb, fa, fm, fb) = bracket
        rho_best = xm
        profit_best, v_best, w_best, growers_best =\
            self.procurer.profitvwg(rho=rho_best) #initialize
        assert np.allclose(-fm,profit_best)
        print """"NOTE: SEARCH for equilibrium, including modern sector.
        Our initial rho value (based on previous rho values) is: {}.
        This rho gives us a profit of {}.
        (We'll try to do better than that.)  BEGIN SEARCH ...
        """.format(rho_best, profit_best)
        f2min = lambda p: -self.procurer.profitvwg(rho=p)[0]
        print """NOTE: starting bracket is ({},{},{});
        starting fvals are ({},{},{})
        """.format(xa,xm,xb,fa,fm,fb)
        #next do the simplest imaginable bracket search for the best rho
        rho, negprofit = simple_bracket_search(f2min, xa, xm, xb, fa, fm, fb)
        profit, v, w, growers = self.procurer.profitvwg(rho=rho)
        #:note: just some error checks
        assert profit == -negprofit, "{} vs {}".format(profit,-negprofit)
        assert profit >= profit_best, "{} vs. {}".format(profit,profit_best) 
        #
        if profit > profit_best:
            profit_best, v_best, w_best, rho_best, growers_best = \
                profit, v, w, rho, growers
        msg = "delta_pct {}: procurer chooses rho={}".format(self.delta_pct,rho_best)
        self.logger.info(msg)
        print(msg)
        return dict(v=v_best
                   ,w=w_best
                   ,rho=rho_best
                   ,growers=growers_best
                   ,profit=profit_best
                   )

 
    def setup_run(self, delta):
        """Return None. Reset the firm and producers.
        Distribute land among producers.
        This prepares for one run (i.e., one land distribution).
        :side-effects: reset firm & producers; distribute land
        """
        #setup firm
        self.procurer.reset()
        #setup producers
        producers = self.producers
        #new land distribution!
        self.set_landholdings(producers, self.delta_pct/100.)
        self.test_runready()

    def set_landholdings(self, producers, delta):
        """Return None, set the hbar and bbar of all `producers`.
        Set the display size of the producers, based on land holdings.

        Parameters
        ----------
        producers (list) : the producers
        delta (float)    : the distribution parameter

        :note: we turn delta into a Gini, and the call `incremental_shares_pareto`
        :called-by: setup_run
        :side-effects: set hbar and bbar and thereby baseInfo for each producer
        """
        if not (delta > 0.0) and (delta < 1.0): #preconditions
            raise ValueError("delta={} but must be in (0,1)".format(delta))
        H, N1 = self.H, self.N1
        params = self.params
        phi, theta = params['phi'], params['theta']
        ishares = incremental_shares_pareto(N1, (1-delta) / (1+delta))
        p = 0  #initialize proportion of landed producers
        for i, prdr in enumerate(producers):  #prdr doesn't need to know its p
            #compute hbar (i.e., land holding)
            if i < N1:
                share = ishares[i]
                hbar = H * share
            else:
                p = 0
                hbar = 0 #prdr is landless
            #Calculate bbar up front instead of each of the many times needed.
            #in principle, landless could access credit (but, phi=0 in EK)
            bbar = phi + (theta * hbar)
            prdr.baseInfo = AgentInfo(hbar=hbar, bbar=bbar, A=prdr._A)
            #display size will depend on land ownership
            if self.has_gui:
                prdr.display(shapesize=(share + 0.2, p + 0.2))

    def clear_factor_markets_traditional(self):
        """Return 2-tuple of float, the equilibrium factor prices
        when there is no modern sector.
        :side-effects: none
        Parameters
        -----------
        add2vwinit : bool
          whether to append to vw_init01 (breaking referential transparency)
        """
        return self.clear_factor_markets(rho=1, growers=())

    def clear_factor_markets(self, rho, growers):
        """Return None. Simultaneous factor market clearing.
        :side-effects: none
        Parameters
        -----------
        rho : float
          relative procurer price
        growers : sequence
          procurers who are offered contracts
        """
        logger = self.logger
        logger.debug("World: Enter clear_factor_markets_simultaneously.")
        #define factor market clearing:
        f2solve = lambda vw: self.xss_land_labor(vw=vw, rho=rho, growers=growers)
        success = False   #we'll set this to True if we get convergence
        if(rho==1):
            vw_init = self.vw_init01
        else:
            assert rho > 1
            vw_init = self.vw_init
        for ct, vw in enumerate(reversed(vw_init)): #try stored starting values
            try: #comment: tighter xtol does not help
                output = fsolve(f2solve, vw, full_output=1, xtol=1e-06)
                (v, w) = map(abs, output[0]) #needed because xss_land_labor uses abs
            except minpack.error: #factor prices should stay positive
                continue  #try another set of initial values
            if output[2]==1:
                (z1,z2) = output[1]['fvec']
                if (abs(z1 <1) and abs(z2 < 1)):
                    success = True
                    break

        #the rest is just reporting with one exception: may append to vw_init
        msg = "Given rho={:6.4f}, ".format(rho)
        if success:
            msg += "factor markets clear at v={:6.4f}, w={:6.4f}.".format(v, w)
            msg += "\n(factor mkt convergence in {} tries)".format(ct+1)
            logger.debug(msg)
            print(msg)
        else:
            msg += '\nfactor market fails to converge:\n' + output[-1]
            logger.warn(msg)
            print(msg)
        self.logger.debug("World: Exit clear_factor_markets_simultaneously.")
        assert (v > 0 and w > 0)
        return v, w

    def producer_distances(self, loc):
        """Return dict, mapping agents to floats.
        The distances of the agents from `loc`.
        Memoized: past computations stored by this.
        :side-effects: none.
        """
        #first check if we have already stored the distances:
        distances = self._producer_distances
        if distances is None:
            producers = self.producers
            distances = dict()
            fx, fy = loc
            for prdr in producers:
                ax, ay = prdr.position
                distances[prdr] = math.hypot(fx-ax, fy-ay)
        else:
            assert loc == self.procurer.position
        return distances


    def producer_infos(self, v, w, rho, growers):
        """Return dict, mapping producers to their info.
        :note: expensive, because `calc_info` calls `choose` for each producer
        :side-effects: none
        """
        result = dict()
        for prdr in self.producers:
            is_grower = prdr in growers
            info = prdr.calc_info(v=v, w=w, rho=rho, has_contract=is_grower)
            contract = info['contract']
            assert is_grower == bool(contract & CONTRACTS['offer'])
            assert info['vwrho'] == (v, w, rho)
            result[prdr] = info
        return result


    def updated_producer_infos(self, infodict):
        """Return dict, mapping producers to properties
        WITH ex post short-side constraints imposed (see below)
        :called-by: `schedule` and `choose_contracts`
        :side-effects: none
        """
        infodict = infodict.copy()  #avoid side effects
        infoseq = infodict.values()
        producers = self.producers
        #We seem usually to succeed in clearing factor markets,
        # but just in case: if not, take the short side.
        #LAND market
        land_supply, land_demand = self.sd_land(infoseq)
        assert (np.allclose(self.H,land_supply) and land_demand >= 0)
        if (land_supply >= land_demand):
            for info in infoseq:
                assert "h" not in info
                info['h'] = info['h_d']
        else: # xss_land < 0, excess demand for land
            scale = land_supply / land_demand
            assert (0 < scale < 1)
            for info in infoseq:
                assert "h" not in info
                #:todo: might be nice to attend to hbar?
                info['h'] = scale * info['h_d']
        #LABOR market
        labor_supply, labor_demand = self.sd_labor(infoseq)
        #print "sd L:", labor_supply, labor_demand
        assert (labor_supply >= 0 and labor_demand >= 0)
        #:note: we will allow for a small approximation error:
        if abs(labor_supply - labor_demand) < 0.01*self.H:
            for info in infoseq:
                assert "L" not in info
                info['L'] = info['L_d']
                info['tw'] = info['tw_d']
        elif labor_supply < labor_demand: #excess demand for labor
            scale = labor_supply / labor_demand
            assert (0 <= scale < 1)
            for info in infoseq:
                assert "L" not in info
                info['tw'] = info['tw_d']
                info['L'] = scale * info['L_d']
        else: #excess supply of labor
            scale = labor_demand / labor_supply
            assert (0 <= scale < 1)
            for info in infoseq:
                assert "L" not in info
                assert "tw" not in info
                info['tw'] = scale * info['tw_d']
                print "scaled:", scale, info['tw_d'], info['tw']
                info['L'] = info['L_d']
        params = self.params
        D, s1, s2 = params['D'], params['s1'], params['s2']
        povertyline = params['povertyline']
        for prdr, info  in infodict.items():
            L = info['L']
            info['ts'] = ts = (s1 * L) + (s2 * L * L)
            info['th'] = th = info['th_d']
            info['tr'] = tr = 1 - info['tw'] - th - ts
            assert ("q" not in info) and ("Y" not in info) and ("U" not in info)
            info['q'], info['Y'], info['U'] = q, Y, U = prdr.qYU(info)
            assert np.allclose(info['U'], info['Y'] + D * math.sqrt(info['tr']) )
            info['is_poor'] = int(Y < povertyline)
        self.logger.debug("World: Exit updated_producer_infos.")
        return infodict

    def xss_land_labor(self, vw, rho, growers):
        """Return 2-tuple of float: the excess supplies.
        Called repeatedly to clear factor markets (given `rho`, `growers`).

        vw : 2-tuple of float
            land rental price, wage
        rho : float
            modern sector price premium
        :side-effects: none
        """
        v, w = vw
        #constrain v>0, w>0
        v, w = abs(v), abs(w)
        infodict = self.producer_infos(v=v, w=w, rho=rho, growers=growers)
        infoseq = tuple(infodict.values())
        xss_land = self.xss_land(infoseq)
        xss_labor = self.xss_labor(infoseq)
        return xss_land, xss_labor

    def xss_land(self, infoseq): 
        """Return float, the excess supply of land.
        :side-effects: none
        """
        land_supply, land_demand = self.sd_land(infoseq)
        return self.xss(land_supply, land_demand)

    def xss_labor(self, infoseq):  
        """Return float, the excess supply of labor.
        :side-effects: none
        """
        labor_supply, labor_demand = self.sd_labor(infoseq)
        return self.xss(labor_supply, labor_demand)

    def sd_land(self, infoseq): 
        """Return float, the excess supply of land.

        Parameters
        ----------
        infoseq : tuple of dict
            producers info for all producers
        :side-effects: none
        """
        land_supply = self.H
        land_demand = sum(info['h_d'] for info in infoseq)
        return (land_supply, land_demand)

    def sd_labor(self, infoseq):  
        """Return float, the excess supply of labor.

        Parameters
        ----------
        infoseq : tuple of dict
            info dict for each producer
        :side-effects: none
        """
        labor_supply = sum(info['tw_d'] for info in infoseq)
        labor_demand = sum(info['L_d'] for info in infoseq)
        return (labor_supply, labor_demand)

    @staticmethod
    def xss(s, d):
        """Return float, the (possibly scaled) value of s-d.
        Checks for non-negativity.
        """
        assert (d >= 0 and s >= 0)
        if (d==0 and s==0):
            msg = "WARN:\n"
            msg += 'supply: {0}; demand: {1}'.format(s, d)
            print(msg)
            logging.warn(msg)
            result = 0
        elif (d==0 or s==0):
            #result = 2*(s-d)/(s+d)
            result = s - d
        else:
            result = s - d
        return result

    def market_info(self, infoseq, v, w):
        """Return str, a market description.
        infoseq : list of dict
          each dict is the info for a producer
        """
        info = "Market info at v={v} and w={w}.".format(v=v,w=w)
        producers = self.producers
        land_supply = sum(prdr.hbar for prdr in producers)
        assert np.allclose(land_supply,self.H)
        land_demand = sum(prdr['h_d'] for prdr in infoseq)
        info += """\nLand Market:
        xs supply:    {hxss}
        (scaled xss:{mhxss})
        land supply:  {hs}
        land demand:  {h_d}
        n suppliers:  {hsupp}
        n demanders:  {hdmnd}
        """.format(
        hs=land_supply,
        h_d=land_demand,
        hxss=land_supply-land_demand,
        mhxss=self.xss(land_supply,land_demand),
        hsupp=sum(prdr.hbar>0 for prdr in producers),
        hdmnd=sum(prdr['h_d']>0 for prdr in infoseq)
        )
        labor_supply = sum(prdr['tw_d'] for prdr in infoseq)
        labor_demand = sum(prdr['L_d'] for prdr in infoseq)
        info += """
        Labor Market:
        xs supply:    {nxss}
        (scaled xss:{mnxss})
        labor supply: {ns}
        labor demand: {nd}
        n suppliers:  {nsupp}
        n demanders:  {ndmnd}
        """.format(
        ns=labor_supply,
        nd=labor_demand,
        nxss=labor_supply-labor_demand,
        mnxss=self.xss(labor_supply,labor_demand),
        nsupp=sum(prdr['tw_d']>0 for prdr in infoseq),
        ndmnd=sum(prdr['L_d']>0 for prdr in infoseq)
        )
        return info

    def setup_outfiles(self):
        """Return None.  Set up the output files.
        There are four core output files:
        `oldmacro` and `oldmicro` record data from
        the model without a modern sector,
        `newmacro` and `newmicro` record data from
        the model with a modern sector.
        Since the filenames are replicate-specific,
        they are set in `get_param_set` (in parameters.py).
        :side-effects: overwrite outfiles with headers
        """
        params = self.params
        self.oldmacro = params['oldmacro']
        self.oldmicro = params['oldmicro']
        self.newmacro = params['newmacro']
        self.newmicro = params['newmicro']

        """Add a header to the output files."""
        with open(self.oldmacro, 'w') as fout:
            fout.write( ','.join(macro_datanames) )
        with open(self.oldmicro, 'w') as fout:
            fout.write( ','.join(micro_datanames) )
        with open(self.newmacro, 'w') as fout:
            fout.write( ','.join(macro_datanames) )
        with open(self.newmicro, 'w') as fout:
            fout.write( ','.join(micro_datanames) )

    def log_macrodata(self, fname, mktinfo, infodict):
        """Return None; log macro data from the simulation.
        :todo: streamline accumulated cruft
        :side-effects: none
        """
        params = self.params
        KM = params['KM']
        KF = params['KF']
        PF = params['PF']
        n_agents = params['n_agents']
        cdistscale = params['cdistscale']
        theta = params['theta']
        povertyline = params['povertyline']
        H = self.H
        #
        delta_pct = mktinfo.delta_pct
        v = mktinfo.v
        w = mktinfo.w
        rho = mktinfo.rho
        growers = mktinfo.growers
        prdrs = list(infodict.keys())
        assert len(prdrs) == n_agents
        assert set(self.producers) == set(prdrs)
        for prdr, info in infodict.items():
            assert (v,w,rho) == info['vwrho'],\
            "v {}, w {}, rho {}, vwrho {}".format(v,w,rho,info['vwrho'])
            Y = info['Y']
            if (Y==0):
                assert (info['K']==0 and info['sector']==0)
        infoseq = infodict.values()
        if (rho==1):
            assert all(info['contract'] == 0 for info in infodict.values())
        #record parameter values
        firm = self.procurer
        delta = delta_pct / 100.
        #class and sector counts
        land_use = defaultdict(float)
        class_counts = defaultdict(int)
        sector_counts = defaultdict(int)
        Ulandless = 0
        n_modern, n_poor = 0, 0
        modern_sector_agents = list()
        n_contract = 0  #number offered contracts
        n_accept = 0  #number accepted contracts
        Qtotal, QMtotal = 0, 0
        Ytotal, YMtotal = 0, 0
        Utotal, UMtotal = 0, 0
        for prdr, info in infodict.items():
            Qtotal += info['q']
            Ytotal += info['Y']
            Utotal += info['U']
            land_use[info['pclass']] += info['h']
            class_counts[info['pclass']] += 1
            sector_counts[info['sector']] += 1
            if prdr.hbar == 0:
                assert info['hbar'] == 0
                Ulandless += info['U']
            if bool(info['contract'] & CONTRACTS['offer']):
                n_contract += 1
            if bool(info['contract'] & CONTRACTS['accept']):
                assert bool(info['contract'] & CONTRACTS['offer'])
                n_accept += 1
            if info['sector'] == 2:
                QMtotal += info['q']  #modern sector output
                YMtotal += info['Y']  #modern sector incomes
                UMtotal += info['U']  #modern sector incomes
                modern_sector_agents.append(prdr)
                n_modern += 1
            n_poor += info['is_poor']
        assert n_agents == sum(class_counts.values())
        Qav = Qtotal / float(n_agents)
        Yav = Ytotal / float(n_agents)
        Uav = Utotal / float(n_agents)
        assert n_modern == len(modern_sector_agents)
        QMav = 0 if (n_modern==0) else QMtotal / float(n_modern)
        YMav = 0 if (n_modern==0) else YMtotal / float(n_modern)
        UMav = 0 if (n_modern==0) else UMtotal / float(n_modern)
        #class counts
        n_class0 = class_counts[0]
        n_class1 = class_counts[1]
        n_class2 = class_counts[2]
        n_class3 = class_counts[3]
        n_class4 = class_counts[4]
        assert n_class0 + n_class1 + n_class2 + n_class3 + n_class4 == n_agents,\
            """Instead of equal values,
            we find {} and {}.""".format(ct,class_counts.values())
        #sector counts
        n_sector0 = sector_counts[0]
        n_sector1 = sector_counts[1]
        n_sector2 = sector_counts[2]
        assert n_sector0 + n_sector1 + n_sector2 == sum(sector_counts.values()) == n_agents,\
            """Instead of equal values,
            we find {} and {}.""".format(ct,sector_counts.values())
        #land use
        assert np.allclose(H, sum(prdr.hbar for prdr in prdrs))
        assert np.allclose(H, sum(info['hbar'] for info in infoseq))
        hsumPL = sum(info['h'] for info in infoseq if info['pclass']==0)
        hsumLC = sum(info['h'] for info in infoseq if info['pclass']==1)
        hsumSC = sum(info['h'] for info in infoseq if info['pclass']==2)
        hsumSM = sum(info['h'] for info in infoseq if info['pclass']==3)
        hsumLG = sum(info['h'] for info in infoseq if info['pclass']==4)
        if params['phi'] <= 0:
            assert np.allclose(0, hsumPL)
        assert np.allclose((hsumPL,hsumLC,hsumSC,hsumSM,hsumLG),
            (land_use[0],land_use[1],land_use[2],land_use[3],land_use[4]))
        if not np.allclose(H, hsumPL+hsumLC+hsumSC+hsumSM+hsumLG):
            msg = "land use = {0} but land supply = {1}"
            msg = msg.format(hsumLC+hsumSC+hsumSM+hsumLG,H)
            self.logger.warn(msg)
        hd_total = sum(info['h_d'] for info in infoseq)
        twd_total = sum(info['tw_d'] for info in infoseq)
        thd_total = sum(info['th_d'] for info in infoseq)
        Ld_total = sum(info['L_d'] for info in infoseq)
        s1, s2 = params['s1'], params['s2']
        tsd_total01 = sum(s1*info['L_d'] for info in infoseq)
        tsd_total02 = sum(s2*info['L_d']**2 for info in infoseq)
        tsd_total = tsd_total01 + tsd_total02
        assert QMtotal == sum(info['q'] for info in infoseq if info['sector']==2)
        p_qmodern = QMtotal/Qtotal
        Ys = list(info['Y'] for info in infoseq)
        Y = sum(Ys)
        gini_y = gini(Ys)
        p_ymodern = YMtotal / Y
        Us = list(info['U'] for info in infodict.values())
        gini_u = gini(Us)
        assert all(U > 0 for U in Us)
        p_poor = float(n_poor) / float(n_agents)   #casts for reader convenience 
        ms_counts = Counter(infodict[a]['pclass'] for a in modern_sector_agents)
        self.logger.info('{0} modern agents, class: {1}'.format(n_modern,ms_counts))
        n_modern = len(modern_sector_agents)
        p_modern = float(n_modern) / float(n_agents)   #casts for reader convenience 
        n_honor = sum(bool(info['contract'] & CONTRACTS['fhonor']) for info in infoseq)
        n_prenege = sum(bool(info['contract'] & CONTRACTS['prenege']) for info in infoseq)
        n_frenege = sum(bool(info['contract'] & CONTRACTS['frenege']) for info in infoseq)
        print "contracts {}, modern {}, frenege {}, prenege {}, accept {}".format(
            n_contract, n_modern, n_frenege, n_prenege, n_accept)
        assert (n_contract==sum(bool(info['contract']) for info in infoseq))
        p_contract = float(n_contract) / float(n_agents)  #cast for reader convenience
        assert rho == firm.rho  #:todo: remove
        Pi = firm.profitxp(rho, infodict)
        if __debug__ and (rho==1.0):
            assert (Pi == 0), "Pi shd be zero (since rho=1) but Pi={}".format(Pi)
        with open(fname, 'a') as fout:
            strdata = [str(eval(x)) for x in macro_datanames]
            fout.write('\n' + ','.join(strdata))

    def log_microdata(self, fname, infodict):
        """Return None.
        Log micro data from the simulation."""
        params = self.params
        attnames = micro_datanames
        assert attnames[0]=='delta_pct'
        with open(fname, 'a') as fout:
            for prdr, info in infodict.items():
                strdata = [str(self.delta_pct)]
                strdata += list(str(info[attr]) for attr in attnames[1:])
                fout.write('\n' + ','.join(strdata))

    def test_setup(self):
        """Return None. Basic tests of overall setup.
        Also see `test_runready`.
        """
        params = self.params
        #agents
        assert len(self.producers) == params['n_agents'], 'shd match'
        #firm
        firm = self.procurer
        assert firm.params is params
        assert firm.world is self

    def test_runready(self):
        """Return None.
        """
        params = self.params
        producers = self.producers
        assert len(producers) == params['n_agents'], 'shd match'
        assert np.allclose(self.H,sum(a.hbar for a in producers)),\
            "H={0}; sum={1}".format(self.H, sum(a.hbar for a in producers))
        landless = [a for a in producers if a.hbar==0]
        assert len(landless)==self.N0, "{} vs {}".format(len(landless),self.N0)
        firm = self.procurer
        assert firm._growers is None
        assert firm._honored is None
        assert firm.rho == 1
        #the next test is a bit expensive and has low payoff; remove? chk
        infodict = self.producer_infos(v=1,w=1,rho=1, growers=())
        assert len(infodict) == params['n_agents'], 'shd match'
        assert all(info['sector']<2 for info in infodict.values())

    #properties

    @property
    def params(self):
        """Return dict, the model parameters.
        """
        return self._params

    @params.setter
    def params(self, prms):
        """Sets the `params` property.
        Used only once per simulation (during setup).
        """
        if self._params is None:
            self._params = prms
        else:
            raise ValueError('params can only be set once')

    #read-only properties

    @property
    def procurer(self):
        """Return Procurer.
        """
        if self._procurer is None:
            self._procurer, = self.get_agents(Procurer)
        return self._procurer

    @property
    def producers(self):
        """Return tuple of Producer.
        """
        if self._producers is None:
            self._producers = tuple(self.get_agents(Producer))
        return self._producers

    #minor safety features

    @property
    def v(self):
        raise NotImplementedError

    @property
    def w(self):
        raise NotImplementedError

    #GUI stuff:

    def redraw_producers(self, infodict):
        """Return None. Redraws producers.
        GUI relevance only.
        """
        if not self.has_gui:
            return
        for prdr, info in infodict.items():
            prdr.redraw(info)


class GUI(gridworld.GridWorldGUI):
    """Provides a GUI for the model run.
    For user convenience only.
    """
    def gui(self):
        """Display buttons and plots."""
        self.add_button('Set Up', 'setup') 
        self.add_button('Run', 'run')   #run calls schedule (see gridworld.py)
        self.add_button('Stop', 'stop')


# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 autoindent
