"""
Create scenarios (parameter dicts) for the simulations.
"""
import ConfigParser
import glob, math, os, random
import numpy as np


class Scenarios(object):
    """Provides a class for representing the baseline
    (as defined in the file `fbaseline`)
    and associated scenarios
    (as defined in the file `fscenarios`).
    :note: output filenames are not part of a scenario
    """
    _baseline = None
    def __init__(self, fbaseline, fscenarios):
        """Return None.
        :side-effects: set `self.baseconfig` and `self.scenarios`
          each to a `ConfigParser`.

        fbaseline : str
          filename for baseline parameters
        fscenarios : str
          filename for scenarios (deviations from baseline)
        """
        # get the baseline parameters (parameter names are the *sections*,
        #  parameter values are the "value" *option*)
        self.baseconfig = baseconfig = ConfigParser.ConfigParser()
        with open("baseline.ini","r") as fin:
            baseconfig.readfp(fin)
        # get the scenarios from `fscenarios` (scenario names are the *sections*;
        #  parameter names are the *options*)
        self.scenarios = scenarios = ConfigParser.ConfigParser()
        scenarios.optionxform = str #otherwise options will be lower-cased
        with open(fscenarios,"r") as fin:
            scenarios.readfp(fin)


    @property
    def baseline(self):
        """Return dict, the baseline scenario (from `fbaseline`).
        """
        if self._baseline is None:
            baseline_parameters = dict(name="baseline")
            baseconfig = self.baseconfig #get parsed fbaseline
            #each baseconfig section is a parameter
            for section in baseconfig.sections():
                #computed values are added after the others are set;
                #currently only grid_side is computed
                if baseconfig.get(section,"value") == "computed":
                    continue
                if baseconfig.get(section,"type") == "float":
                    val = baseconfig.getfloat(section,"value") 
                elif baseconfig.get(section,"type") == "int":
                    val = baseconfig.getint(section,"value")
                elif baseconfig.get(section,"type") == "bool":
                    val = baseconfig.getboolean(section,"value")
                else: #string default
                    val = baseconfig.get(section,"value")
                baseline_parameters[section] = val
            self.update_computed_values(baseline_parameters)
            self._baseline = baseline_parameters
        return self._baseline

    def update_computed_values(self, params):
        """Return None.
        :note: currently only grid_side is computed
        :side-effects: mutates params
        """
        n_agents = params["n_agents"]
        params["grid_side"] = int(round(2 * math.sqrt(n_agents)))

    def get_scenario(self, name):
        """Return dict, the parameter configuration for one scenario,
        as specified first in `fbaseline` and then in `fscenarios`.
        The scenario specifications in `fscenarios` are only
        parameter values that deviate from the baseline.
        Thus we being with (a copy of) the baseline, and modify it.

        name : str
          the name of the scenario (i.e., the section in `fscenarios`)
        :raises: AssertionError
        """
        scenario = self.baseline.copy()
        if (name == "baseline"):
            return scenario
        scenario["name"] = name
        scenarios = self.scenarios
        assert name in scenarios.sections(), \
            "{} not available in scenarios.ini".format(name)
        for param, val in scenarios.items(name):
            assert param in scenario, \
                "{} not in scenario; shd only be *re*setting values".format(param)
            val = self.cast_config_val(param, val) #cast val to specified type
            scenario[param] = val
        self.update_computed_values(scenario) #reset grid_side to match scenario
        return scenario

    def cast_config_val(self, param, val):
        """Return obj, of type determined by `fbaseline`.
        param : str
          the parameter name
        val : str
          a string representation of the value
        """
        basecfg = self.baseconfig
        if basecfg.get(param,"type") == "float":
            val = float(val)
        elif basecfg.get(param,"type") == "int":
            val = int(val)
        elif basecfg.get(param,"type") == "bool":
            val = False if (val.lower()=="false") else True
        else: #string default
            val = basecfg.get(experiment, param)
        return val



scenarios = Scenarios("baseline.ini", "scenarios.ini")
baseline = scenarios.baseline
snames = """theta075 theta125 theta200
km050 km060 km065 km090 km100 km125 km150 km150new km225 km500
km085 km095 km070
km150p700
p175 p700
p150 p200 p250 p300 p400 p450
kf075 kf100 kf150 kf200 kf400
na010km200
na010 na025 na050 na075 na100 na125 na500 na750 na1500 na3000
rho125""".split()
for name in snames:
    exec("{name} = scenarios.get_scenario('{name}')".format(name=name))






"""
    #chkchk firm will consider 100 (arbitrary) random locations
    @property
    def outfile03(self):
        return os.path.normpath(os.path.join('out',self.name,'xss.csv'))
    @property
    def outfile04(self):
        return os.path.normpath(os.path.join('out',self.name,'profit.csv'))
    @property
    def logfile(self):
        return os.path.normpath(os.path.join('out',self.name,'logfile.txt'))
"""



#just testing that naming is working as expected
for name in snames:
    assert name == eval(name)['name']


def get_param_set(params, n_sim, overwrite=True):
    """Yield dict, each a replicate specific set of parameters.
    There are `n_sim` (e.g., 100) replicates for the parameters `params`.
    Each replicate is given its own random seed, sim_num, and outfile names.
    This generator function should yield dicts with **pickleable** elements only,
    so that it can be used with multiprocessing.
    
    params : dict
      the scenario to be simulated
    n_sim : int
      the number of replicates
    overwrite : bool
      whether or not to overwrite previous results
    """
    name = params['name']
    #all output goes below `out` folder, which should already exist
    outdir = os.path.join("out",name)
    if not os.path.exists(outdir):
        print "creating output directory"
        os.mkdir(outdir) 
    #"old" and "new" means without and with modern sector
    outfiles = ('oldmacro', 'newmacro', 'oldmicro', 'newmicro')
    previously_done = list()
    if not overwrite:
        for outname in outfiles:
            outfile_pattern = outname + '???.csv'
            search_pattern = os.path.normpath(os.path.join(outdir, outfile_pattern))
            previously_done.extend(glob.glob(search_pattern))
    previously_done = set(int(x[-7:-4]) for x in previously_done) #just the replicate numbers
    for n in range(n_sim):
        if n in previously_done:
            print 'Skip: simulation {0} (already completed).'.format(n)
            continue
        else:
            simparams = params.copy()  #fresh copy for each replicate (IMPORTANT!)
            simparams['sim_num'] = n
            seed = (params['rndseed'], n)   #replicate(n)-specific seed
            simparams['prng'] = random.Random(seed)
            loggername =  "{0}replicate{1:03d}".format(name,n) #replicate specific logger name
            simparams['loggername'] = loggername
            simparams['logfile'] = os.path.normpath(os.path.join(outdir, loggername+".txt"))
            for outfile in outfiles: #create replication specific output files
                fullname = "{0}{1:03d}.csv".format(outfile, n)
                simparams[outfile] = os.path.normpath(os.path.join(outdir, fullname))
            yield simparams

def get_lhs_paramset(baseparams, limitdict, nsamples, overwrite=True, prng=None):
    """Yield dict, a simulation specific set of parameters.
    There will be len(limitdict) dicts of parameters `params`.
    Each replicate has its own random seed, sim_num, and outfile names.
    This generator function should yield dicts with **pickleable** elements only,
    so that it can be used with multiprocessing.
    """
    if prng is None:
        prng = np.random.RandomState(314) #used by `lhs`
    exp_name = baseparams['name']
    outdir = os.path.join('out',exp_name)
    outfiles = ('oldmacro', 'newmacro', 'oldmicro', 'newmicro')
    """
    previously_done = list()
    for outname in outfiles:
        outfile_pattern = outname + '???.csv'
        search_pattern = os.path.normpath(os.path.join(outdir, outfile_pattern))
        previously_done.extend(glob.glob(search_pattern))
    if not overwrite:
        previously_done = set(int(x[-7:-4]) for x in previously_done)
    elif previously_done:
        doremove = raw_input("Remove files from last experiment of {}?(y,n)".format(exp_name))
        if doremove[0].lower()=='y':
            for fname in previously_done:
                os.remove(fname)
            previously_done = []
    """
    pnames, plims = zip(*limitdict.items())
    design = lhs(nsamples, plims, midpt=False, prng=prng)
    for n, row in enumerate(design):
        simparams = baseparams.copy()  #fresh copy for each sim (IMPORTANT!)
        simparams.update(zip(pnames,row))
        simparams['n_agents'] = int(round(simparams['n_agents'])) #restore int value
        simparams['sim_num'] = n
        seed = (baseparams['rndseed'], n)   #replicate(n)-specific seed
        simparams['prng'] = random.Random(seed)
        simparams['logfile'] = os.path.normpath(os.path.join(outdir, 'logfile.txt'))
        simparams['loggername'] = '{0}{1:03d}'.format(exp_name,n)
        for outfile in outfiles:
            fullname = "{0}{1:03d}.csv".format(outfile, n)
            simparams[outfile] = os.path.normpath(os.path.join(outdir, fullname))
        yield simparams



#for each param, divide range into n intervals, and use each interval *once*
def lhsdesign(n, p, midpt=False, prng=None):
    """Return array. Each of the `n` rows
    has `p` values.

    Parameters
    ----------
    n : int
      number of samples
    p : int
      the number of "parameters"
    midpt : bool
      Use midpoints if true. Else sample randomly
      from each interval.
    prng : RandomState
      A NumPy random number generator.
    """
    if prng is None:
        prng = np.random.RandomState()
    tpdesign = np.empty((p,n), dtype=float)
    for idx in range(p):
        #left bound
        smpl = np.arange(n, dtype=float)
        #add portion of interval
        if midpt:
            smpl += 0.5
        else:
            smpl += prng.random_sample(n)
        #scale
        smpl /= float(n)
        #shffle
        prng.shuffle(smpl)
        tpdesign[idx] = smpl
    return tpdesign.transpose()

def lhs(n, lohiseq, midpt=False, prng=None):
    """Return array. Each of the `n` rows
    has a value for each of the (len(lohiseq) parameters,
    whose ranges are provided by lohiseq.  

    Parameters
    ----------
    n : int
      number of samples
    lohiseq : sequence of 2 tuples
      the (low,high) range for each parameter
    midpt : bool
      Use midpoints if True. Else sample randomly
      from each interval.
    prng : RandomState
      A NumPy random number generator.
    """
    if prng is None:
        prng = np.random.RandomState()
    p = len(lohiseq)
    design = lhsdesign(n, p, midpt, prng)
    for idx, row in enumerate(lohiseq):
        low, high = row
        delta = (high-low)
        #scale
        design[:,idx] *= delta
        #translate
        design[:,idx] += low
    return design

if __name__ == '__main__':
    for p in get_param_set(bracket, 2):
        print p['oldmicro']
    exit()
    for p in get_param_set(oldbaseline_smallsearch, 2):
        print p['oldmicro']
    for p in get_lhs_paramset(baseline, dict(n_agents=(100,500)), 10):
        print p['newmacro']
        print p['n_agents']

# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 autoindent
